# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Functions to anonymize user data and activity."""


from copy import deepcopy

from flask import current_app
from invenio_accounts.models import LoginInformation, SessionActivity, User, userrole
from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from invenio_circulation.proxies import current_circulation
from invenio_db import db
from invenio_oauthclient.models import RemoteAccount, RemoteToken, UserIdentity
from invenio_search.engine import search as inv_search
from invenio_userprofiles.models import UserProfile

from invenio_app_ils.acquisition.api import ORDER_PID_TYPE
from invenio_app_ils.acquisition.proxies import current_ils_acq
from invenio_app_ils.circulation.search import (
    get_active_loans_by_patron_pid,
    get_loans_by_patron_pid,
)
from invenio_app_ils.document_requests.api import DOCUMENT_REQUEST_PID_TYPE
from invenio_app_ils.errors import AnonymizationActiveLoansError, PatronNotFoundError
from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE
from invenio_app_ils.ill.proxies import current_ils_ill
from invenio_app_ils.notifications.models import NotificationsLogs
from invenio_app_ils.patrons.api import get_patron_or_unknown_dump
from invenio_app_ils.proxies import current_app_ils


def get_patron_activity(patron_pid):
    """Get activity related to the given patron pid."""
    if patron_pid is None:
        raise ValueError("No patron pid was provided.")

    patron = get_patron_or_unknown_dump(patron_pid)

    def dump(search):
        return [hit.to_dict() for hit in search.scan()]

    DocumentRequestSearch = current_app_ils.document_request_search_cls
    patron_document_requests = dump(
        DocumentRequestSearch().search_by_patron_pid(patron_pid)
    )

    BorrowingRequestsSearch = current_ils_ill.borrowing_request_search_cls
    patron_borrowing_requests = dump(
        BorrowingRequestsSearch().search_by_patron_pid(patron_pid)
    )

    OrderSearch = current_ils_acq.order_search_cls
    patron_acquisitions = dump(OrderSearch().search_by_patron_pid(patron_pid))

    patron_loans = dump(get_loans_by_patron_pid(patron_pid))

    patron_profile = UserProfile.get_by_userid(patron_pid).__dict__
    del patron_profile["_sa_instance_state"]

    patron_data = {
        "patron": patron,
        "profile": patron_profile,
        "document_requests": patron_document_requests,
        "borrowing_requests": patron_borrowing_requests,
        "acquisitions": patron_acquisitions,
        "loans": patron_loans,
    }

    return patron_data


def anonymize_patron_data(patron_pid, force=False):
    """Anonymize all the data/activity related to the given patron pid."""
    if patron_pid is None:
        raise ValueError("No patron pid was provided.")

    SystemAgent = current_app.config["ILS_PATRON_SYSTEM_AGENT_CLASS"]
    if str(patron_pid) == str(SystemAgent.id):
        raise ValueError("The patron pid cannot be the SystemAgent")

    try:
        patron = current_app_ils.patron_cls.get_patron(patron_pid)
    except PatronNotFoundError:
        patron = None
        if not force:
            raise PatronNotFoundError(patron_pid)

    n_loans = get_active_loans_by_patron_pid(patron_pid).count()
    if n_loans > 0:
        raise AnonymizationActiveLoansError(
            "Cannot delete user {0}: found {1} active loans.".format(
                patron_pid, n_loans
            )
        )
    OrderSearch = current_ils_acq.order_search_cls
    n_orders = OrderSearch().get_ongoing_orders_by_patron_pid(patron_pid).count()
    if n_orders > 0:
        raise AnonymizationActiveLoansError(
            "Cannot delete user {0}: found {1} active orders.".format(
                patron_pid, n_orders
            )
        )

    # get anonymous patron values
    cls = current_app.config["ILS_PATRON_ANONYMOUS_CLASS"]
    anonymous_patron_fields = cls().dumps_loader()

    Loan = current_circulation.loan_record_cls
    BorrowingRequest = current_ils_ill.borrowing_request_record_cls
    DocumentRequest = current_app_ils.document_request_record_cls
    Order = current_ils_acq.order_record_cls

    anonymized_records = {
        CIRCULATION_LOAN_PID_TYPE: {
            "indexer": current_circulation.loan_indexer(),
            "records": [],
        },
        BORROWING_REQUEST_PID_TYPE: {
            "indexer": current_ils_ill.borrowing_request_indexer_cls(),
            "records": [],
        },
        DOCUMENT_REQUEST_PID_TYPE: {
            "indexer": current_app_ils.document_request_indexer,
            "records": [],
        },
        ORDER_PID_TYPE: {
            "indexer": current_ils_acq.order_indexer,
            "records": [],
        },
    }

    patron_loans = get_loans_by_patron_pid(patron_pid).scan()

    indices = 0
    for hit in patron_loans:
        loan = Loan.get_record_by_pid(hit.pid)

        completed = (
            current_app.config["CIRCULATION_STATES_LOAN_CANCELLED"]
            + current_app.config["CIRCULATION_STATES_LOAN_COMPLETED"]
        )
        if loan["state"] not in completed:
            params = deepcopy(loan)
            params.update(
                dict(
                    cancel_reason="Loan request cancelled by the system.",
                    transaction_user_pid=str(SystemAgent.id),
                )
            )
            loan = current_circulation.circulation.trigger(
                loan, **dict(params, trigger="cancel")
            )
        loan["patron_pid"] = anonymous_patron_fields["pid"]
        loan["patron"] = anonymous_patron_fields
        loan.commit()
        anonymized_records[CIRCULATION_LOAN_PID_TYPE]["records"].append(loan)
        indices += 1

    BorrowingRequestsSearch = current_ils_ill.borrowing_request_search_cls
    patron_borrowing_requests = (
        BorrowingRequestsSearch().search_by_patron_pid(patron_pid).scan()
    )

    for hit in patron_borrowing_requests:
        borrowing_request = BorrowingRequest.get_record_by_pid(hit.pid)
        borrowing_request["patron"] = anonymous_patron_fields
        borrowing_request["patron_pid"] = anonymous_patron_fields["pid"]
        borrowing_request.commit()
        anonymized_records[BORROWING_REQUEST_PID_TYPE]["records"].append(borrowing_request)
        indices += 1

    DocumentRequestSearch = current_app_ils.document_request_search_cls
    patron_document_requests = (
        DocumentRequestSearch().search_by_patron_pid(patron_pid).scan()
    )

    for hit in patron_document_requests:
        document_request = DocumentRequest.get_record_by_pid(hit.pid)
        if document_request["state"] == "PENDING":
            document_request["state"] = "DECLINED"
            document_request["decline_reason"] = "USER_CANCEL"
        document_request["patron"] = anonymous_patron_fields
        document_request["patron_pid"] = anonymous_patron_fields["pid"]
        document_request.commit()
        anonymized_records[ORDER_PID_TYPE]["records"].append(document_request)
        indices += 1

    patron_acquisitions = OrderSearch().search_by_patron_pid(patron_pid).scan()

    for hit in patron_acquisitions:
        acquisition = Order.get_record_by_pid(hit.pid)
        for line in acquisition["order_lines"]:
            if line.get("patron_pid") == patron_pid:
                line["patron_pid"] = anonymous_patron_fields["pid"]
        acquisition.commit()
        anonymized_records[Order._pid_type]["records"].append(acquisition)
        indices += 1

    # delete rows from db
    dropped = delete_user_account(patron_pid)

    # anonymize recipient id in notifications
    notifications = anonymize_patron_in_notification_logs(patron_pid)

    db.session.commit()

    # index all after committing to DB, to ensure that no errors occurred.
    for value in anonymized_records.values():
        indexer, records = value["indexer"], value["records"]
        for record in records:
            indexer.index(record)

    if patron:
        try:
            patron_indexer = current_app_ils.patron_indexer
            patron_indexer.delete(patron)
        except inv_search.NotFoundError:
            pass

    return dropped, indices, notifications


def delete_user_account(patron_pid):
    """Deletes a user account from the database, without updating the index."""
    dropped = 0

    with db.session.begin_nested():
        d = db.session.query(userrole).filter(userrole.c.user_id == patron_pid)
        dropped += d.delete(synchronize_session=False) or 0

        dropped += (
            SessionActivity.query.filter(SessionActivity.user_id == patron_pid).delete()
            or 0
        )

        dropped += (
            UserIdentity.query.filter(UserIdentity.id_user == patron_pid).delete() or 0
        )

        ra = RemoteAccount.query.filter(
            RemoteAccount.user_id == patron_pid
        ).one_or_none()
        if ra:
            dropped += (
                RemoteToken.query.filter(
                    RemoteToken.id_remote_account == ra.id
                ).delete()
                or 0
            )
            dropped += ra.delete() or 0

        dropped += (
            UserProfile.query.filter(UserProfile.user_id == patron_pid).delete() or 0
        )
        dropped += (
            LoginInformation.query.filter(
                LoginInformation.user_id == patron_pid
            ).delete()
            or 0
        )
        dropped += User.query.filter(User.id == patron_pid).delete() or 0

    return dropped


def anonymize_patron_in_notification_logs(patron_pid):
    """Anonymizes patron in notification log db."""
    notification_anonymizations = 0
    AnonymousPatron = current_app.config["ILS_PATRON_ANONYMOUS_CLASS"]

    with db.session.begin_nested():
        notifications = NotificationsLogs.query.filter_by(
            recipient_user_id=patron_pid
        ).all()
        for notification in notifications:
            notification.recipient_user_id = AnonymousPatron.id
            notification_anonymizations += 1
    db.session.commit()

    return notification_anonymizations



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS Records views."""

from flask import Blueprint, abort, current_app, request
from flask_login import current_user
from invenio_files_rest.models import ObjectVersion
from invenio_records_rest.utils import obj_or_import_string
from invenio_records_rest.views import pass_record
from invenio_rest import ContentNegotiatedMethodView

from invenio_app_ils.documents.api import DOCUMENT_PID_TYPE
from invenio_app_ils.eitems.api import EITEM_PID_TYPE
from invenio_app_ils.errors import StatsError
from invenio_app_ils.permissions import backoffice_permission
from invenio_app_ils.records.permissions import RecordPermission
from invenio_app_ils.series.api import SERIES_PID_TYPE
from invenio_app_ils.signals import file_downloaded, record_viewed


def create_document_stats_blueprint(app):
    """Create document stats blueprint."""
    blueprint = Blueprint("ils_document_stats", __name__, url_prefix="")
    endpoints = app.config.get("RECORDS_REST_ENDPOINTS", [])

    def register_view(pid_type):
        options = endpoints.get(pid_type, {})
        default_media_type = options.get("default_media_type", "")
        rec_serializers = options.get("record_serializers", {})
        serializers = {
            mime: obj_or_import_string(func) for mime, func in rec_serializers.items()
        }

        stats_view = DocumentStatsResource.as_view(
            DocumentStatsResource.view_name.format(pid_type),
            serializers=serializers,
            default_media_type=default_media_type,
        )
        blueprint.add_url_rule(
            "{0}/stats".format(options["item_route"]),
            view_func=stats_view,
            methods=["POST"],
        )

    register_view(DOCUMENT_PID_TYPE)
    register_view(SERIES_PID_TYPE)
    register_view(EITEM_PID_TYPE)
    return blueprint


class DocumentStatsResource(ContentNegotiatedMethodView):
    """Document stats resource."""

    view_name = "{}_stats"

    @pass_record
    def post(self, pid, record, **kwargs):
        """Send a signal to count record view for the record stats."""
        factory = RecordPermission(record, "read")
        if not factory.is_public() and not backoffice_permission().can():
            if not current_user.is_authenticated:
                abort(401)
            abort(403)
        data = request.get_json()
        event_name = data.get("event")
        if event_name == "record-view":
            record_viewed.send(
                current_app._get_current_object(),
                pid=pid,
                record=record,
            )
            return self.make_response(pid, record, 202)
        elif event_name == "file-download":
            if "key" not in data:
                abort(406, "File key is required")
            if "bucket_id" not in record:
                abort(406, "Record has no bucket")
            obj = ObjectVersion.get(record["bucket_id"], data["key"])
            file_downloaded.send(
                current_app._get_current_object(), obj=obj, record=record
            )
            return self.make_response(pid, record, 202)
        return StatsError(
            description="Invalid stats event request: {}".format(event_name)
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Literature search module."""

from invenio_search.api import DefaultFilter, RecordsSearch

from invenio_app_ils.search_permissions import (
    ils_search_factory,
    search_filter_record_permissions,
)


def search_factory_literature(self, search, query_parser=None):
    """Search factory for literature (series and documents)."""

    def filter_serial_issues(search, query_string=None):
        """Filter periodical issues unless include_all is specified."""
        from distutils.util import strtobool

        from flask import request

        include_all = request.values.get("include_all", "no")
        if include_all == "":
            include_all = "yes"

        if not strtobool(include_all):
            issue_query_string = "NOT document_type:SERIAL_ISSUE"
            if query_string:
                query_string = "({}) AND ({})".format(query_string, issue_query_string)
            else:
                query_string = issue_query_string
        return search, query_string

    return ils_search_factory(self, search, query_parser, filter_serial_issues)


class LiteratureSearch(RecordsSearch):
    """Literature search that searches both documents and series."""

    boosted_fields = [
        "identifiers.value^12.0",
        "identifiers.value.text^12.0",
        "title^8.0",
        "authors.full_name^6.0",
        "imprint.publisher^4.0",
        "edition^4.0",
        "keywords^2.0",
        "abstract^2.0",
    ]

    class Meta:
        """Search for documents and series."""

        index = ["documents", "series"]
        doc_types = None
        default_filter = DefaultFilter(search_filter_record_permissions)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test loans item permissions."""

import json
from copy import deepcopy

from flask import url_for

from tests.helpers import user_login

NEW_LOAN = {
    "item_pid": {"type": "pitmid", "value": "itemid-1"},
    "document_pid": "docid-1",
    "patron_pid": "1",
    "transaction_location_pid": "locid-1",
    "pickup_location_pid": "locid-1",
}


def _fetch_loan(client, json_headers, loan):
    """Return the loan fetched with a REST call."""
    url = url_for("invenio_records_rest.loanid_item", pid_value=loan["pid"])
    return client.get(url, headers=json_headers)


def _assert_get_loan_success(client, json_headers, loan):
    """Assert that GET API call succeeded."""
    res = _fetch_loan(client, json_headers, loan)
    assert res.status_code == 200
    data = json.loads(res.data.decode("utf-8"))["metadata"]
    assert data["pid"] == loan["pid"]
    assert data["patron_pid"] == loan["patron_pid"]


def test_anonymous_cannot_get_any_loan(client, json_headers, testdata):
    """Test that anonymous cannot get any loan."""
    loan_patron1 = testdata["loans"][0]
    assert loan_patron1["patron_pid"] == "1"
    loan_patron2 = testdata["loans"][3]
    assert loan_patron2["patron_pid"] == "2"

    res = _fetch_loan(client, json_headers, loan_patron1)
    assert res.status_code == 401
    res = _fetch_loan(client, json_headers, loan_patron2)
    assert res.status_code == 401


def test_admin_or_librarian_can_get_any_loan(client, json_headers, users, testdata):
    """Test that admins and librarians can get any loan."""
    for username in ["admin", "librarian"]:
        user_login(client, username, users)
        # GET loans of Patron 1
        loan_patron1 = testdata["loans"][0]
        assert loan_patron1["patron_pid"] == "1"
        _assert_get_loan_success(client, json_headers, loan_patron1)

        # GET loans of Patron 2
        loan_patron2 = testdata["loans"][3]
        assert loan_patron2["patron_pid"] == "2"
        _assert_get_loan_success(client, json_headers, loan_patron2)


def test_patron_can_get_only_his_loans(client, json_headers, users, testdata):
    """Test that patrons can get only their loans."""
    loan_patron1 = testdata["loans"][0]
    assert loan_patron1["patron_pid"] == "1"

    loan_patron2 = testdata["loans"][3]
    assert loan_patron2["patron_pid"] == "2"

    # Patron 1 GET his loans
    user_login(client, "patron1", users)
    _assert_get_loan_success(client, json_headers, loan_patron1)
    # Patron 1 GET loans of Patron 2
    res = _fetch_loan(client, json_headers, loan_patron2)
    assert res.status_code == 403

    # Patron 2 GET his loans
    user_login(client, "patron2", users)
    _assert_get_loan_success(client, json_headers, loan_patron2)
    # Patron 2 GET loans of Patron 1
    res = _fetch_loan(client, json_headers, loan_patron1)
    assert res.status_code == 403


def _test_post_new_loan(client, json_headers, user_id, response_code):
    """Test POST new loan."""
    url = url_for("invenio_records_rest.loanid_list")
    params = deepcopy(NEW_LOAN)
    params["transaction_user_pid"] = user_id
    res = client.post(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == response_code
    if res.status_code == 201:
        data = json.loads(res.data.decode("utf-8"))["metadata"]
        return data["pid"]


def _test_replace_existing_loan(client, json_headers, user_id, loanid, response_code):
    """Test PUT existing loan."""
    url = url_for("invenio_records_rest.loanid_item", pid_value=loanid)
    params = deepcopy(NEW_LOAN)
    params["transaction_user_pid"] = user_id
    res = client.put(url, headers=json_headers, data=json.dumps(params))
    assert res.status_code == response_code


def _test_delete_existing_loan(client, json_headers, loanid, response_code):
    """Test DELETE existing loan."""
    url = url_for("invenio_records_rest.loanid_item", pid_value=loanid)
    res = client.delete(url, headers=json_headers)
    assert res.status_code == response_code


def test_anonymous_cannot_update_loans(client, json_headers, testdata):
    """Test that anonymous cannot update loans."""
    loanid = "loanid-1"
    _test_post_new_loan(client, json_headers, user_id="", response_code=401)
    _test_replace_existing_loan(
        client, json_headers, user_id="", loanid=loanid, response_code=401
    )


def test_patron_cannot_update_loans(client, json_headers, users, testdata):
    """Test that patrons cannot update loans."""
    loanid = "loanid-1"
    user = user_login(client, "patron1", users)
    _test_post_new_loan(client, json_headers, user.id, 403)
    _test_replace_existing_loan(client, json_headers, user.id, loanid, 403)
    _test_delete_existing_loan(client, json_headers, loanid, 403)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS Eitems APIs."""

from functools import partial

from flask import current_app
from invenio_files_rest.models import ObjectVersion
from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2

from invenio_app_ils.errors import DocumentNotFoundError
from invenio_app_ils.fetchers import pid_fetcher
from invenio_app_ils.minters import pid_minter
from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.records.api import IlsRecord, RecordValidator

EITEM_PID_TYPE = "eitmid"
EITEM_PID_MINTER = "eitmid"
EITEM_PID_FETCHER = "eitmid"

EItemIdProvider = type(
    "EItemIdProvider",
    (RecordIdProviderV2,),
    dict(pid_type=EITEM_PID_TYPE, default_status=PIDStatus.REGISTERED),
)
eitem_pid_minter = partial(pid_minter, provider_cls=EItemIdProvider)
eitem_pid_fetcher = partial(pid_fetcher, provider_cls=EItemIdProvider)


class EItemValidator(RecordValidator):
    """EItem record validator."""

    def ensure_document_exists(self, document_pid):
        """Ensure document exists or raise."""
        Document = current_app_ils.document_record_cls
        try:
            Document.get_record_by_pid(document_pid)
        except PIDDoesNotExistError:
            raise DocumentNotFoundError(document_pid)

    def validate(self, record, **kwargs):
        """Validate record before create and commit."""
        super().validate(record, **kwargs)

        document_pid = record["document_pid"]

        self.ensure_document_exists(document_pid)


class EItem(IlsRecord):
    """EItem record class."""

    EITEM_TYPES = [
        "E-BOOK",
        "AUDIOBOOK",
        "VIDEO",
    ]

    _pid_type = EITEM_PID_TYPE
    _schema = "eitems/eitem-v3.0.0.json"
    _document_resolver_path = (
        "{scheme}://{host}/api/resolver/eitems/{eitem_pid}/document"
    )
    _files_resolver_path = "{scheme}://{host}/api/resolver/eitems/{eitem_pid}/files"
    _validator = EItemValidator()

    @classmethod
    def get_document_pid(cls, eitem_pid):
        """Retrieve the referenced document PID of the given item PID."""
        eitem = cls.get_record_by_pid(eitem_pid)
        return eitem["document_pid"]

    @classmethod
    def build_resolver_fields(cls, data):
        """Build all resolver fields."""
        data["document"] = {
            "$ref": cls._document_resolver_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                eitem_pid=data["pid"],
            )
        }
        data["files"] = {
            "$ref": cls._files_resolver_path.format(
                scheme=current_app.config["JSONSCHEMAS_URL_SCHEME"],
                host=current_app.config["JSONSCHEMAS_HOST"],
                eitem_pid=data["pid"],
            )
        }

    @classmethod
    def create(cls, data, id_=None, **kwargs):
        """Create EItem record."""
        cls.build_resolver_fields(data)
        return super().create(data, id_=id_, **kwargs)

    def update(self, *args, **kwargs):
        """Update EItem record."""
        super().update(*args, **kwargs)
        self.build_resolver_fields(self)

    @property
    def files(self):
        """Get EItem files."""
        if "bucket_id" not in self:
            return []
        return ObjectVersion.get_by_bucket(self["bucket_id"]).limit(1000).all()


def eitem_event_builder(event, sender_app, obj=None, **kwargs):
    """Add eitem metadata to the event."""
    record = kwargs["record"]
    event.update(dict(eitem_pid=record["pid"], document_pid=record.get("document_pid")))
    return event



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Ils permissions."""

from functools import wraps

from flask import abort, current_app, g, has_request_context
from flask_login import current_user
from flask_principal import UserNeed
from invenio_access import action_factory
from invenio_access.permissions import Permission, authenticated_user, superuser_access
from invenio_records_rest.utils import allow_all, deny_all

from invenio_app_ils.errors import InvalidLoanExtendError, LoanCheckoutByPatronForbidden
from invenio_app_ils.proxies import current_app_ils

backoffice_access_action = action_factory("ils-backoffice-access")


def need_permissions(action):
    """View decorator to check permissions for the given action or abort.

    :param action: The action needed.
    """

    def decorator_builder(f):
        @wraps(f)
        def decorate(*args, **kwargs):
            check_permission(
                current_app.config["ILS_VIEWS_PERMISSIONS_FACTORY"](action)
            )
            return f(*args, **kwargs)

        return decorate

    return decorator_builder


def check_permission(permission):
    """Abort if permission is not allowed.

    :param permission: The permission to check.
    """
    if permission is not None and not permission.can():
        if not current_user.is_authenticated:
            abort(401)
        abort(403)


def authenticated_user_permission(*args, **kwargs):
    """Return an object that evaluates if the current user is authenticated."""
    return Permission(authenticated_user)


def backoffice_permission(*args, **kwargs):
    """Return permission to allow only librarians and admins."""
    return Permission(backoffice_access_action)


def superuser_permission(*args, **kwargs):
    """Return permission to allow only admins."""
    return Permission(superuser_access)


def patron_permission(patron_pid):
    """Return a permission for the given patron."""
    return Permission(UserNeed(int(patron_pid)), backoffice_access_action)


def file_download_permission(obj):
    """File download permissions."""
    bucket_id = str(obj.bucket_id)
    search_cls = current_app_ils.eitem_search_cls
    results = search_cls().search_by_bucket_id(bucket_id)
    if len(results) != 1:
        return deny_all()

    eitem_cls = current_app_ils.eitem_record_cls
    record = eitem_cls.get_record_by_pid(results[0].pid)
    if record.get("open_access", False):
        return allow_all()
    return authenticated_user_permission()


def files_permission(obj, action=None):
    """Return permission for Files REST."""
    if action == "object-read":
        return file_download_permission(obj)
    return backoffice_permission()


def loan_extend_circulation_permission(loan):
    """Return permission to allow only owner and librarians to extend loan."""
    if not has_request_context():
        # not in a request context, probably from CLI
        return allow_all()

    if current_user.is_anonymous:
        abort(401)

    is_admin_or_librarian = backoffice_permission().allows(g.identity)
    is_logged_in_patron_owner = current_user.id == int(loan["patron_pid"])
    if not is_admin_or_librarian and is_logged_in_patron_owner:
        Document = current_app_ils.document_record_cls
        document_rec = Document.get_record_by_pid(loan["document_pid"])
        document = document_rec.replace_refs()
        is_overbooked = document.get("circulation", {}).get("overbooked")
        if is_overbooked is None:
            # NOTE: this should never happen, if it happens it means that the
            # document does not have `circulation.overbooked` field. Fix it!
            abort(500)
        elif is_overbooked:
            raise InvalidLoanExtendError(
                "The extension cannot be automatically accepted due to high "
                "demand for this literature. Please contact the library to "
                "request a loan extension."
            )
    return PatronOwnerPermission(loan)


def patron_owner_permission(record):
    """Return permission to allow owner and librarian to access the record."""
    if not has_request_context():
        # allows performing the actions out of the request context f.e. CLI
        return allow_all()
    return PatronOwnerPermission(record)


def loan_checkout_permission(*args, **kwargs):
    """Loan checkout permissions checks.

    Allow admins and librarians to checkout, patrons to self-checkout when enabled.
    """
    if not has_request_context():
        # CLI or Celery task
        return backoffice_permission()

    if current_user.is_anonymous:
        abort(401)

    is_admin_or_librarian = backoffice_permission().allows(g.identity)
    if is_admin_or_librarian:
        return backoffice_permission()

    # ensure that only the loan's patron can do operations on this loan
    if len(args):
        loan = args[0]
    else:
        loan = kwargs["record"]
    is_patron_current_user = current_user.id == int(loan["patron_pid"])

    if current_app.config["ILS_SELF_CHECKOUT_ENABLED"] and is_patron_current_user:
        return authenticated_user_permission()

    raise LoanCheckoutByPatronForbidden(int(loan["patron_pid"]), current_user.id)


class PatronOwnerPermission(Permission):
    """Return Permission to evaluate if the current user owns the record."""

    def __init__(self, record):
        """Constructor."""
        super().__init__(UserNeed(int(record["patron_pid"])), backoffice_access_action)


_is_authenticated_user = [
    "circulation-loan-request",
    "patron-loans",
    "bulk-loan-extension",
]
_is_backoffice_permission = [
    "circulation-loan-force-checkout",
    "circulation-overdue-loan-notification",
    "circulation-loan-update-dates",
    "relations-create",
    "relations-delete",
    "stats-most-loaned",
    "document-request-actions",
    "bucket-create",
    "ill-brwreq-patron-loan-create",
    "ill-brwreq-patron-loan-extension-accept",
    "ill-brwreq-patron-loan-extension-decline",
    "send-notification-to-patron",
]
_is_patron_owner_permission = [
    "document-request-decline",
    "ill-brwreq-patron-loan-extension-request",
]


def views_permissions_factory(action):
    """Return ILS views permissions factory."""
    if action in _is_authenticated_user:
        return authenticated_user_permission()
    elif action in _is_backoffice_permission:
        return backoffice_permission()
    elif action in _is_patron_owner_permission:
        return PatronOwnerPermission
    elif action == "circulation-loan-checkout":
        return backoffice_permission()
    elif (
        action == "circulation-loan-self-checkout"
        and current_app.config["ILS_SELF_CHECKOUT_ENABLED"]
    ):
        return authenticated_user_permission()
    return deny_all()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test record CRUD."""

import pytest

from invenio_app_ils.errors import DocumentNotFoundError, PatronNotFoundError
from invenio_app_ils.ill.api import BorrowingRequest
from invenio_app_ils.ill.errors import ILLError
from invenio_app_ils.providers.errors import ProviderNotFoundError


def test_brw_req_refs(app, testdata):
    """Test creation of a borrowing request."""
    brw_req = BorrowingRequest.create(
        dict(
            pid="illlid-99",
            document_pid="docid-1",
            patron_pid="1",
            provider_pid="ill-provid-1",
            status="PENDING",
            type="PHYSICAL_COPY",
        )
    )
    assert "$schema" in brw_req
    assert "provider" in brw_req and "$ref" in brw_req["provider"]
    assert "document" in brw_req and "$ref" in brw_req["document"]
    assert "patron" in brw_req and "$ref" in brw_req["patron"]

    brw_req = BorrowingRequest.get_record_by_pid("illbid-1")
    brw_req = brw_req.replace_refs()
    assert "provider" in brw_req and brw_req["provider"]["name"]
    assert "document" in brw_req and brw_req["document"]["title"]
    assert "patron" in brw_req and brw_req["patron"]["name"]


def test_ill_brw_validation(db, testdata):
    """Test validation when updating a borrowing request."""
    borrowing_request_pid = testdata["ill_brw_reqs"][0]["pid"]

    # change document pid
    borrowing_request = BorrowingRequest.get_record_by_pid(borrowing_request_pid)
    borrowing_request["document_pid"] = "not_found_doc"
    with pytest.raises(DocumentNotFoundError):
        borrowing_request.commit()

    # change provider pid
    borrowing_request = BorrowingRequest.get_record_by_pid(borrowing_request_pid)
    borrowing_request["provider_pid"] = "not_found_lib"
    with pytest.raises(ProviderNotFoundError):
        borrowing_request.commit()

    # change patron pid
    borrowing_request = BorrowingRequest.get_record_by_pid(borrowing_request_pid)
    borrowing_request["patron_pid"] = "9999"
    with pytest.raises(PatronNotFoundError):
        borrowing_request.commit()

    # add cancel reason without status CANCELLED
    borrowing_request = BorrowingRequest.get_record_by_pid(borrowing_request_pid)
    borrowing_request["cancel_reason"] = "USER_CANCEL"
    with pytest.raises(ILLError):
        borrowing_request.commit()

    # update status to CANCELLED without cancel_reason
    borrowing_request = BorrowingRequest.get_record_by_pid(borrowing_request_pid)
    borrowing_request["status"] = "CANCELLED"
    with pytest.raises(ILLError):
        borrowing_request.commit()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test records relations."""

import json

from flask import url_for
from invenio_db import db

from invenio_app_ils.signals import record_viewed
from tests.helpers import user_login, user_logout


def _most_loaned_request(client, json_headers, from_date=None, to_date=None):
    """Perform a stats request."""
    params = []
    if from_date is not None:
        params.append("from_date={}".format(from_date))
    if to_date is not None:
        params.append("to_date={}".format(to_date))
    response = client.get(
        "{}?{}".format(
            url_for("invenio_app_ils_circulation_stats.most-loaned"),
            "&".join(params),
        ),
        headers=json_headers,
    )
    return json.loads(response.data.decode("utf-8"))


def _assert_most_loaned(client, json_headers, from_date, to_date, expect):
    """Assert most loaned request."""
    resp = _most_loaned_request(client, json_headers, from_date, to_date)
    hits = resp["hits"]["hits"]
    assert len(hits) == len(expect)
    for hit in hits:
        pid = hit["metadata"]["pid"]
        assert hit["metadata"]["loan_count"] == expect[pid]["loans"]
        assert hit["metadata"]["loan_extensions"] == expect[pid]["extensions"]


def test_stats_most_loaned_documents(client, json_headers, testdata_most_loaned, users):
    """Test most loaned documents API endpoint."""
    user_login(client, "librarian", users)

    # Dates covering all loans
    _assert_most_loaned(
        client,
        json_headers,
        "2019-01-01",
        "2019-12-01",
        expect={
            "docid-1": dict(loans=3, extensions=0),
            "docid-2": dict(loans=1, extensions=1),
            "docid-3": dict(loans=2, extensions=6),
            "docid-5": dict(loans=1, extensions=0),
        },
    )
    # Test checking range which should be empty
    _assert_most_loaned(client, json_headers, "2019-01-01", "2019-01-01", expect={})
    # Test range only including the first loan
    _assert_most_loaned(
        client,
        json_headers,
        "2019-01-01",
        "2019-01-03",
        expect={"docid-1": dict(loans=1, extensions=0)},
    )
    _assert_most_loaned(
        client,
        json_headers,
        "2019-02-02",
        "2019-03-02",
        expect={
            "docid-1": dict(loans=2, extensions=0),
            "docid-2": dict(loans=1, extensions=1),
            "docid-3": dict(loans=1, extensions=3),
        },
    )
    _assert_most_loaned(
        client,
        json_headers,
        "2019-05-20",
        "2019-08-22",
        expect={"docid-3": dict(loans=1, extensions=3)},
    )
    # outside end range
    _assert_most_loaned(client, json_headers, "2019-05-21", "2019-12-31", expect={})


def test_stats_downloads_views_permissions(
    client, json_headers, users, testdata, mocker
):
    """Test permissions on downloads/views stats."""

    # mock record_viewed signal because travis timed out and the signal does
    # not need to be tested here
    mocker.patch(
        "invenio_app_ils.signals.record_viewed.send",
    )

    url_open_access = url_for(
        "ils_document_stats.docid_stats", pid_value="docid-open-access"
    )
    url_closed_access = url_for(
        "ils_document_stats.docid_stats", pid_value="docid-closed-access"
    )
    params = {}
    params["event"] = "record-view"

    # permission for librarian
    user_login(client, "librarian", users)
    res = client.post(url_open_access, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202

    res = client.post(url_closed_access, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202

    user_logout(client)

    # permission for patron
    user_login(client, "patron1", users)

    res = client.post(url_open_access, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202

    res = client.post(url_closed_access, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 403

    user_logout(client)

    # permission for unauthenticated user
    res = client.post(url_open_access, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 202

    res = client.post(url_closed_access, headers=json_headers, data=json.dumps(params))
    assert res.status_code == 401



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests orders CRUD."""

from copy import deepcopy
from datetime import datetime

import pytest

from invenio_app_ils.acquisition.api import Order
from invenio_app_ils.acquisition.errors import AcquisitionError
from invenio_app_ils.errors import DocumentNotFoundError, PatronNotFoundError
from invenio_app_ils.providers.errors import ProviderNotFoundError


def _assert_extra_fields(order):
    """Test that extra fields are automatically added."""
    assert "$schema" in order
    assert "provider" in order and "$ref" in order["provider"]
    assert "resolved_order_lines" in order and "$ref" in order["resolved_order_lines"]


def _assert_refs_values(order):
    """Test that refs are replaced correctly"""
    assert "provider" in order and order["provider"]["name"]
    for order_line in order["resolved_order_lines"]:
        assert "document" in order_line and order_line["document"]["title"]
        assert "patron" in order_line and order_line["patron"]["name"]


def test_order_refs(app, testdata):
    """Test creation of an order."""
    order = Order.create(
        dict(
            pid="acqoid-99",
            order_date=datetime.now().isoformat(),
            status="ORDERED",
            provider_pid="acq-provid-1",
            order_lines=[
                dict(
                    document_pid="docid-3",
                    recipient="PATRON",
                    patron_pid="1",
                    medium="PAPER",
                    copies_ordered=2,
                )
            ],
        )
    )
    _assert_extra_fields(order)

    order = Order.get_record_by_pid("acqoid-6")
    order = order.replace_refs()
    _assert_refs_values(order)


def test_acq_order_validation(db, testdata):
    """Test validation when updating an order."""
    order_pid = testdata["acq_orders"][0]["pid"]
    orginal_order = Order.get_record_by_pid(order_pid)

    # add cancel reason without status CANCELLED
    order = deepcopy(orginal_order)
    order["cancel_reason"] = "A reason to cancel"
    with pytest.raises(AcquisitionError):
        order.commit()

    # update status to CANCELLED without cancel_reason
    order = deepcopy(orginal_order)
    order["status"] = "CANCELLED"
    with pytest.raises(AcquisitionError):
        order.commit()

    # change provider pid
    order = deepcopy(orginal_order)
    order["provider_pid"] = "not-existing"
    with pytest.raises(ProviderNotFoundError):
        order.commit()

    order = deepcopy(orginal_order)
    first_order_line = order["order_lines"][0]

    # change document pid
    first_order_line["document_pid"] = "not_found_doc"
    with pytest.raises(DocumentNotFoundError):
        order.commit()

    order = deepcopy(orginal_order)
    first_order_line = order["order_lines"][0]

    # change patron pid
    first_order_line["patron_pid"] = "99999"
    with pytest.raises(PatronNotFoundError):
        order.commit()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App Provider loaders."""

from invenio_app_ils.records.loaders import ils_marshmallow_loader

from .jsonschemas.provider import ProviderSchemaV1

provider_loader = ils_marshmallow_loader(ProviderSchemaV1)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS circulation stats views."""

from datetime import datetime

from flask import Blueprint, current_app, request
from invenio_pidstore import current_pidstore
from invenio_records_rest.utils import obj_or_import_string
from invenio_rest import ContentNegotiatedMethodView

from invenio_app_ils.circulation.stats.api import fetch_most_loaned_documents
from invenio_app_ils.config import RECORDS_REST_MAX_RESULT_WINDOW
from invenio_app_ils.documents.api import DOCUMENT_PID_FETCHER, DOCUMENT_PID_TYPE
from invenio_app_ils.errors import InvalidParameterError
from invenio_app_ils.permissions import need_permissions


def create_most_loaned_documents_view(blueprint, app):
    """Add url rule for most loaned documents."""
    endpoints = app.config.get("RECORDS_REST_ENDPOINTS", [])
    document_endpoint = endpoints.get(DOCUMENT_PID_TYPE, {})
    default_media_type = document_endpoint.get("default_media_type", "")
    search_serializers_aliases = document_endpoint.get("search_serializers_aliases", "")
    search_serializers = document_endpoint.get("search_serializers")
    serializers = {
        mime: obj_or_import_string(func) for mime, func in search_serializers.items()
    }

    view_class = MostLoanedDocumentsResource

    blueprint.add_url_rule(
        "/circulation/stats/most-loaned",
        view_func=view_class.as_view(
            view_class.view_name,
            serializers=serializers,
            serializers_query_aliases=search_serializers_aliases,
            default_media_type=default_media_type,
        ),
        methods=["GET"],
    )


def create_circulation_stats_blueprint(app):
    """Add statistics views to the blueprint."""
    blueprint = Blueprint("invenio_app_ils_circulation_stats", __name__, url_prefix="")

    create_most_loaned_documents_view(blueprint, app)

    return blueprint


class MostLoanedDocumentsResource(ContentNegotiatedMethodView):
    """Statistics view for the documents with the most loans."""

    view_name = "most-loaned"
    default_bucket_size = 30

    def __init__(self, *args, **kwargs):
        """Constructor."""
        super().__init__(*args, **kwargs)
        endpoints = current_app.config.get("RECORDS_REST_ENDPOINTS", [])
        document_endpoint = endpoints.get(DOCUMENT_PID_TYPE, {})
        self.max_result_window = document_endpoint.get(
            "max_result_window", RECORDS_REST_MAX_RESULT_WINDOW
        )

    def _validate_bucket_size(self):
        """Validate bucket size parameter."""
        size_param = request.args.get("size", self.default_bucket_size)
        try:
            value = int(size_param)
        except ValueError:
            msg = "Parameter `size` is not a number: {}".format(size_param)
            raise InvalidParameterError(description=msg)

        if value > self.max_result_window:
            msg = "Parameter `size` should be lower than {}".format(
                self.max_result_window
            )
            raise InvalidParameterError(description=msg)
        return value

    def _validate_start_date_range(self):
        """Validate start date range parameters."""

        def validate_date(param, date):
            """Validate a date."""
            try:
                return datetime.strptime(date, "%Y-%m-%d")
            except ValueError:
                msg = "Parameter '{}' is invalid: {}".format(param, date)
                raise InvalidParameterError(description=msg)

        from_date = request.args.get("from_date", None)
        from_date_obj = None
        to_date = request.args.get("to_date", None)
        to_date_obj = None

        if from_date:
            from_date_obj = validate_date("from_date", from_date)
        if to_date:
            to_date_obj = validate_date("to_date", to_date)

        if from_date_obj and to_date_obj and to_date_obj < from_date_obj:
            msg = "Parameter 'to_date' cannot be before 'from_date'."
            raise InvalidParameterError(description=msg)

        return from_date, to_date

    @need_permissions("stats-most-loaned")
    def get(self, *args, **kwargs):
        """Get the most loaned documents and include the extension count.

        Params:
            {
                from_date: start date,
                to_date: end date,
                [size: maximum number of documents to return (buckets)]
            }
        """
        size = self._validate_bucket_size()
        from_date, to_date = self._validate_start_date_range()
        most_loaned_documents = fetch_most_loaned_documents(from_date, to_date, size)
        return self.make_response(
            pid_fetcher=current_pidstore.fetchers[DOCUMENT_PID_FETCHER],
            search_result=most_loaned_documents,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS Locations search APIs."""

from invenio_search.api import RecordsSearch


class LocationSearch(RecordsSearch):
    """RecordsSearch for locations."""

    class Meta:
        """Search only on locations index."""

        index = "locations"
        doc_types = None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the Item referenced in the Document."""

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.eitems.api import EItem
from invenio_app_ils.proxies import current_app_ils

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred Items for a Document record."""
    from flask import current_app

    def stock_resolver(document_pid):
        """Search and return the mediums of the document."""
        item_search = current_app_ils.item_search_cls()
        eitem_search = current_app_ils.eitem_search_cls()

        search = item_search.search_by_document_pid(document_pid)
        search.aggs.bucket("mediums", "terms", field="medium")
        search_response = search.execute()
        mediums = [
            bucket.key for bucket in search_response.aggregations.mediums.buckets
        ]
        eitems = eitem_search.search_by_document_pid(document_pid)
        for type in EItem.EITEM_TYPES:
            type_count = eitems.filter("term", eitem_type=type).count()
            if type_count > 0:
                mediums.append(type)
        return {
            "mediums": mediums,
        }

    url_map.add(
        Rule(
            "/api/resolver/documents/<document_pid>/stock",
            endpoint=stock_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the circulation status referenced in the Document."""

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.circulation.search import (
    get_loan_next_available_date,
    get_loans_aggregated_by_states,
    get_overdue_loans_by_doc_pid,
)
from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE
from invenio_app_ils.items.search import get_items_aggregated_by_statuses

# Note: there must be only one resolver per file,
# otherwise only the last one is registered


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred circulation aggregation for a Document record."""
    from flask import current_app

    def circulation_resolver(document_pid):
        """Return circulation info for the given Document."""
        # loans
        loan_states = (
            current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
            + current_app.config["CIRCULATION_STATES_LOAN_COMPLETED"]
            + current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]
        )

        past_loans_count = 0
        active_loans_count = 0
        pending_loans_count = 0
        loans_search = get_loans_aggregated_by_states(document_pid, loan_states)

        loans_result = loans_search.execute()

        for bucket in loans_result.aggregations.states.buckets:
            if bucket["key"] in current_app.config["CIRCULATION_STATES_LOAN_COMPLETED"]:
                past_loans_count += bucket["doc_count"]
            elif bucket["key"] in current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]:
                active_loans_count += bucket["doc_count"]
            elif bucket["key"] in current_app.config["CIRCULATION_STATES_LOAN_REQUEST"]:
                pending_loans_count += bucket["doc_count"]

        overdue_loans_count = get_overdue_loans_by_doc_pid(document_pid).count()

        # items
        unavailable_items_count = 0
        items_for_reference_count = 0
        item_statuses = get_items_aggregated_by_statuses(document_pid)
        item_result = item_statuses.execute()
        items_count = item_result.hits.total.value
        for bucket in item_result.aggregations.statuses.buckets:
            if bucket["key"] not in "CAN_CIRCULATE":
                unavailable_items_count += bucket["doc_count"]
            if bucket["key"] in "FOR_REFERENCE_ONLY":
                items_for_reference_count = bucket["doc_count"]

        active_ill_loans_count = 0

        for hit in loans_result.hits:
            loan = hit.to_dict()
            loan_item_type = loan.get("item_pid", {}).get("type")
            is_brw_req = loan_item_type == BORROWING_REQUEST_PID_TYPE
            if (
                loan["state"] in current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
                and is_brw_req
            ):
                active_ill_loans_count += 1

        active_ils_loans_count = active_loans_count - active_ill_loans_count

        available_items_for_loan_count = (
            items_count - active_ils_loans_count - unavailable_items_count
        )

        circulation = {
            "active_loans_count": active_loans_count,
            "available_items_for_loan_count": available_items_for_loan_count,
            "can_circulate_items_count": items_count - unavailable_items_count,
            "items_for_reference_count": items_for_reference_count,
            "overbooked": pending_loans_count > available_items_for_loan_count,
            "overdue_loans_count": overdue_loans_count,
            "past_loans_count": past_loans_count,
            "pending_loans_count": pending_loans_count,
        }

        if (
            circulation["overbooked"]
            or circulation["active_loans_count"]
            >= circulation["available_items_for_loan_count"]
        ):
            next_available_loans = (
                get_loan_next_available_date(document_pid).execute().hits
            )
            total = next_available_loans.total.value
            if total > 0:
                next_date = next_available_loans[0].end_date
                circulation["next_available_date"] = next_date
        return circulation

    url_map.add(
        Rule(
            "/api/resolver/documents/<document_pid>/circulation",
            endpoint=circulation_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test notifications permissions."""

from flask import url_for

from tests.helpers import user_login

_HTTP_OK = [200, 201, 204]
_HTTP_UNAUTHORIZED = [401]
_HTTP_FORBIDDEN = [403]
_HTTP_NOT_FOUND = [404]
MAIL_ID = 1
ITEM_ENDPOINT = "invenio_app_ils_notifications.get_notification"
LIST_ENDPOINT = "invenio_app_ils_notifications.get_notifications"


def test_notifications_read_permissions(client, json_headers, users):
    """Test notifications endpoints read permissions."""
    list_tests = [
        ("admin", _HTTP_OK),
        ("librarian", _HTTP_OK),
        ("patron1", _HTTP_FORBIDDEN),
        ("anonymous", _HTTP_UNAUTHORIZED),
    ]

    item_tests = [
        ("admin", _HTTP_NOT_FOUND),
        ("librarian", _HTTP_NOT_FOUND),
        ("patron1", _HTTP_FORBIDDEN),
        ("anonymous", _HTTP_UNAUTHORIZED),
    ]

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_read(expected_status, id):
        """Test record read."""
        url = url_for(ITEM_ENDPOINT, id=id)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, expected_status in list_tests:
        user_login(client, username, users)
        _test_list(expected_status)

    for username, expected_status in item_tests:
        user_login(client, username, users)
        _test_read(expected_status, MAIL_ID)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test vocabularies endpoint."""
import json

from flask import url_for

from tests.helpers import user_login

LIST_ENDPOINT = "invenio_records_rest.vocid_list"


def test_vocabularies_permissions(client, testdata, json_headers, users):
    """Test vocabularies endpoints permissions."""
    tests = [
        ("admin", [200]),
        ("librarian", [200]),
        ("patron1", [200]),
        ("anonymous", [200]),
    ]

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, codes in tests:
        user_login(client, username, users)
        _test_list(codes)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Circulation notifications APIs."""

from flask import current_app
from invenio_records_rest.utils import obj_or_import_string

from invenio_app_ils.notifications.api import build_common_msg_ctx, send_notification
from invenio_app_ils.patrons.api import Patron


def _build_circulation_msg_ctx():
    """Return a dict with extra circulation ctx to pass to jinja templates."""
    return dict(
        circulation_delivery_methods=current_app.config[
            "ILS_CIRCULATION_DELIVERY_METHODS"
        ]
    )


def send_loan_notification(loan, action, msg_extra_ctx=None, **kwargs):
    """Send notification to patron about a loan action."""
    _filter = current_app.config["ILS_CIRCULATION_NOTIFICATIONS_FILTER"]
    if _filter and _filter(loan, action, **kwargs) is False:
        return

    msg_ctx = msg_extra_ctx or {}
    msg_ctx.update(build_common_msg_ctx(loan))
    msg_ctx.update(_build_circulation_msg_ctx())

    func_or_path = current_app.config["ILS_CIRCULATION_NOTIFICATIONS_MSG_BUILDER"]
    builder = obj_or_import_string(func_or_path)
    msg = builder(loan, action, msg_ctx, **kwargs)

    patron = msg_ctx["patron"]
    patrons = [patron]

    send_notification(patrons, msg, **kwargs)


def send_loan_overdue_reminder_notification(
    loan, days_ago, is_manually_triggered=False
):
    """Send loan overdue notification."""
    send_loan_notification(
        action="overdue_reminder",
        loan=loan,
        is_manually_triggered=is_manually_triggered,
        msg_extra_ctx=dict(days_ago=days_ago),
    )


def send_expiring_loan_reminder_notification(loan, expiring_in_days):
    """Send reminder notification."""
    send_loan_notification(
        action="expiring_reminder",
        loan=loan,
        msg_extra_ctx=dict(expiring_in_days=expiring_in_days),
    )


def send_bulk_extend_notification(
    extended_loans,
    not_extended_loans,
    patron_pid,
    action="bulk_extend",
    msg_extra_ctx=None,
    **kwargs
):
    """Send notification to patron about bulk extend action."""
    _filter = current_app.config["ILS_CIRCULATION_NOTIFICATIONS_FILTER"]
    if _filter and _filter({}, action, **kwargs) is False:
        return

    msg_ctx = msg_extra_ctx or {}
    msg_ctx["patron"] = Patron.get_patron(patron_pid)
    extended_loans_ctx, not_extended_loans_ctx = [], []
    for loan in extended_loans:
        context = {"loan": loan}
        context.update(build_common_msg_ctx(loan))
        extended_loans_ctx.append(context)
    for loan in not_extended_loans:
        context = {"loan": loan}
        context.update(build_common_msg_ctx(loan))
        not_extended_loans_ctx.append(context)

    msg_ctx.update(
        extended_loans=extended_loans_ctx, not_extended_loans_ctx=not_extended_loans_ctx
    )
    msg_ctx.update(_build_circulation_msg_ctx())

    func_or_path = current_app.config["ILS_CIRCULATION_NOTIFICATIONS_MSG_BUILDER"]
    builder = obj_or_import_string(func_or_path)
    msg = builder({}, action, msg_ctx, **kwargs)

    patron = msg_ctx["patron"]
    patrons = [patron]

    send_notification(patrons, msg, **kwargs)


def circulation_filter_notifications(record, action, **kwargs):
    """Filter notifications to be sent."""
    return True



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test records relations siblings."""

import json

from flask import url_for

from invenio_app_ils.documents.api import Document
from tests.helpers import get_test_record, user_login

from .helpers import (
    recrel_assert_record_relations,
    recrel_choose_endpoints_and_do_request,
)


def _test_sibl_language_relation(client, json_headers):
    """Test creation/deletion siblings language relations."""
    first_pid_value = "docid-1"
    first_pid_type = "docid"
    second_pid_value = "docid-2"
    second_pid_type = "docid"
    third_pid_value = "docid-6"
    third_pid_type = "docid"
    relation_type = "language"

    payload = [
        {
            "pid_value": second_pid_value,
            "pid_type": second_pid_type,
            "relation_type": relation_type,
        },
        {
            "pid_value": third_pid_value,
            "pid_type": third_pid_type,
            "relation_type": relation_type,
        },
    ]

    def _test_create():
        """Test relation creation."""
        rec1, rec2 = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                first_pid_value,
                first_pid_type,
                second_pid_value,
                second_pid_type,
            ),
            payload,
        )
        rec3 = Document.get_record_by_pid(third_pid_value)
        rec3 = rec3.replace_refs()
        recrel_assert_record_relations(
            rec1,
            expected={
                "relations": {
                    "language": [
                        {
                            "pid_value": second_pid_value,
                            "pid_type": second_pid_type,
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec2["title"],
                                "languages": rec2["languages"],
                                "document_type": rec2["document_type"],
                                "publication_year": rec2["publication_year"],
                            },
                        },
                        {
                            "pid_value": third_pid_value,
                            "pid_type": third_pid_type,
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec3["title"],
                                "document_type": rec3["document_type"],
                                "languages": rec3["languages"],
                                "publication_year": rec3["publication_year"],
                            },
                        },
                    ]
                }
            },
        )
        recrel_assert_record_relations(
            rec2,
            expected={
                "relations": {
                    "language": [
                        {
                            "pid_value": first_pid_value,
                            "pid_type": first_pid_type,
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec1["title"],
                                "languages": rec1["languages"],
                                "edition": rec1["edition"],
                                "document_type": rec1["document_type"],
                                "publication_year": rec1["publication_year"],
                            },
                        },
                        {
                            "pid_value": third_pid_value,
                            "pid_type": third_pid_type,
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec3["title"],
                                "languages": rec3["languages"],
                                "document_type": rec3["document_type"],
                                "publication_year": rec3["publication_year"],
                            },
                        },
                    ]
                }
            },
        )
        recrel_assert_record_relations(
            rec3,
            expected={
                "relations": {
                    "language": [
                        {
                            "pid_value": first_pid_value,
                            "pid_type": first_pid_type,
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec1["title"],
                                "languages": rec1["languages"],
                                "edition": rec1["edition"],
                                "document_type": rec1["document_type"],
                                "publication_year": rec1["publication_year"],
                            },
                        },
                        {
                            "pid_value": second_pid_value,
                            "pid_type": second_pid_type,
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec2["title"],
                                "languages": rec2["languages"],
                                "document_type": rec2["document_type"],
                                "publication_year": rec2["publication_year"],
                            },
                        },
                    ]
                }
            },
        )

    def _test_delete():
        """Test relation deletion."""
        rec1, rec2 = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "DELETE"),
            (
                first_pid_value,
                first_pid_type,
                second_pid_value,
                second_pid_type,
            ),
            payload,
        )
        rec3 = Document.get_record_by_pid(third_pid_value)
        rec3 = rec3.replace_refs()
        recrel_assert_record_relations(rec1, expected={"relations": {}})
        recrel_assert_record_relations(rec2, expected={"relations": {}})
        recrel_assert_record_relations(rec3, expected={"relations": {}})

    _test_create()
    _test_delete()
    # recreate for the next one, to have some more valuable test data
    _test_create()


def _test_sibl_edition_relation(client, json_headers, testdata):
    """Test creation/deletion siblings edition relations."""
    first_pid_value = "docid-3"
    first_pid_type = "docid"
    second_pid_value = "docid-1"
    second_pid_type = "docid"
    third_pid_value = "docid-8"
    third_pid_type = "docid"
    relation_type = "edition"

    payload = [
        {
            "pid_value": second_pid_value,
            "pid_type": second_pid_type,
            "relation_type": relation_type,
        },
        {
            "pid_value": third_pid_value,
            "pid_type": third_pid_type,
            "relation_type": relation_type,
        },
    ]

    def _test_create():
        """Test relation creation."""
        rec1, rec2 = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                first_pid_value,
                first_pid_type,
                second_pid_value,
                second_pid_type,
            ),
            payload,
        )
        rec3 = Document.get_record_by_pid(third_pid_value)
        rec3 = rec3.replace_refs()

        recrel_assert_record_relations(
            rec1,
            expected={
                "relations": {
                    "edition": [
                        {
                            "pid_value": second_pid_value,
                            "pid_type": second_pid_type,
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec2["title"],
                                "edition": rec2["edition"],
                                "languages": rec2["languages"],
                                "document_type": rec2["document_type"],
                                "publication_year": rec2["publication_year"],
                            },
                        },
                        {
                            "pid_value": third_pid_value,
                            "pid_type": third_pid_type,
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec3["title"],
                                "edition": rec3["edition"],
                                "document_type": rec3["document_type"],
                                "publication_year": rec3["publication_year"],
                            },
                        },
                    ]
                }
            },
        )
        recrel_assert_record_relations(
            rec3,
            expected={
                "relations": {
                    "edition": [
                        {
                            "pid_value": second_pid_value,
                            "pid_type": second_pid_type,
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec2["title"],
                                "edition": rec2["edition"],
                                "languages": rec2["languages"],
                                "document_type": rec2["document_type"],
                                "publication_year": rec2["publication_year"],
                            },
                        },
                        {
                            "pid_value": first_pid_value,
                            "pid_type": first_pid_type,
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec1["title"],
                                "edition": rec1["edition"],
                                "document_type": rec1["document_type"],
                                "publication_year": rec1["publication_year"],
                            },
                        },
                    ]
                }
            },
        )

        rec_docid_2 = get_test_record(testdata, "documents", "docid-2")
        rec_docid_6 = get_test_record(testdata, "documents", "docid-6")

        recrel_assert_record_relations(
            rec2,
            {
                "relations": {
                    "edition": [
                        {
                            "pid_value": third_pid_value,
                            "pid_type": third_pid_type,
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec3["title"],
                                "edition": rec3["edition"],
                                "document_type": rec3["document_type"],
                                "publication_year": rec3["publication_year"],
                            },
                        },
                        {
                            "pid_value": first_pid_value,
                            "pid_type": first_pid_type,
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec1["title"],
                                "edition": rec1["edition"],
                                "document_type": rec1["document_type"],
                                "publication_year": rec1["publication_year"],
                            },
                        },
                    ],
                    "language": [
                        {
                            "pid_value": rec_docid_2["pid"],
                            "pid_type": "docid",
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec_docid_2["title"],
                                "languages": rec_docid_2["languages"],
                                "document_type": rec_docid_2["document_type"],
                                "publication_year": rec_docid_2["publication_year"],
                            },
                        },
                        {
                            "pid_value": rec_docid_6["pid"],
                            "pid_type": "docid",
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec_docid_6["title"],
                                "document_type": rec_docid_6["document_type"],
                                "languages": rec_docid_6["languages"],
                                "publication_year": rec_docid_6["publication_year"],
                            },
                        },
                    ],
                }
            },
        )

    def _test_delete():
        """Test relation deletion."""
        rec1, rec2 = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "DELETE"),
            (
                first_pid_value,
                first_pid_type,
                second_pid_value,
                second_pid_type,
            ),
            payload,
        )
        rec3 = Document.get_record_by_pid(third_pid_value)
        rec3 = rec3.replace_refs()

        recrel_assert_record_relations(rec1, expected={"relations": {}})
        recrel_assert_record_relations(rec3, expected={"relations": {}})

        rec_docid_2 = get_test_record(testdata, "documents", "docid-2")
        rec_docid_6 = get_test_record(testdata, "documents", "docid-6")

        # docid-1 should still have the languages relation created in the previous test
        recrel_assert_record_relations(
            rec2,
            expected={
                "relations": {
                    "language": [
                        {
                            "pid_value": rec_docid_2["pid"],
                            "pid_type": "docid",
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec_docid_2["title"],
                                "languages": rec_docid_2["languages"],
                                "document_type": rec_docid_2["document_type"],
                                "publication_year": rec_docid_2["publication_year"],
                            },
                        },
                        {
                            "pid_value": rec_docid_6["pid"],
                            "pid_type": "docid",
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec_docid_6["title"],
                                "document_type": rec_docid_6["document_type"],
                                "languages": rec_docid_6["languages"],
                                "publication_year": rec_docid_6["publication_year"],
                            },
                        },
                    ]
                }
            },
        )

    def _test_empty_edition_field():
        edition_first_pid_value = "docid-11"
        edition_first_pid_type = "docid"
        edition_second_pid_value = "docid-12"
        edition_second_pid_type = "docid"

        create_payload = {
            "pid_value": edition_second_pid_value,
            "pid_type": edition_second_pid_type,
            "relation_type": relation_type,
        }

        rec1, rec2 = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                edition_first_pid_value,
                edition_first_pid_type,
                edition_second_pid_value,
                edition_second_pid_type,
            ),
            create_payload,
            expect_status_code=400,
        )

    _test_create()
    _test_delete()
    # recreate for the next one, to have some more valuable test data
    _test_create()
    _test_empty_edition_field()


def _test_sibl_other_relation(client, json_headers, testdata):
    """Test creation/deletion siblings other relations."""
    first_pid_value = "docid-2"
    first_pid_type = "docid"
    second_pid_value = "docid-3"
    second_pid_type = "docid"
    relation_type = "other"

    payload = {
        "pid_value": second_pid_value,
        "pid_type": second_pid_type,
        "relation_type": relation_type,
        "note": "exercise",
    }

    def _test_create():
        """Test relation creation."""
        rec1, rec2 = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                first_pid_value,
                first_pid_type,
                second_pid_value,
                second_pid_type,
            ),
            payload,
        )

        rec_docid_1 = get_test_record(testdata, "documents", "docid-1")
        rec_docid_6 = get_test_record(testdata, "documents", "docid-6")
        rec_docid_8 = get_test_record(testdata, "documents", "docid-8")

        recrel_assert_record_relations(
            rec1,
            expected={
                "relations_extra_metadata": {
                    "other": [
                        {
                            "pid_value": second_pid_value,
                            "pid_type": second_pid_type,
                            "note": "exercise",
                        }
                    ]
                },
                "relations": {
                    "language": [
                        {
                            "pid_value": rec_docid_1["pid"],
                            "pid_type": "docid",
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec_docid_1["title"],
                                "edition": rec_docid_1["edition"],
                                "languages": rec_docid_1["languages"],
                                "document_type": rec_docid_1["document_type"],
                                "publication_year": rec_docid_1["publication_year"],
                            },
                        },
                        {
                            "pid_value": rec_docid_6["pid"],
                            "pid_type": "docid",
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec_docid_6["title"],
                                "document_type": rec_docid_6["document_type"],
                                "languages": rec_docid_6["languages"],
                                "publication_year": rec_docid_6["publication_year"],
                            },
                        },
                    ],
                    "other": [
                        {
                            "pid_value": second_pid_value,
                            "pid_type": second_pid_type,
                            "note": "exercise",
                            "relation_type": "other",
                            "record_metadata": {
                                "title": rec2["title"],
                                "edition": rec2["edition"],
                                "document_type": rec2["document_type"],
                                "publication_year": rec2["publication_year"],
                            },
                        }
                    ],
                },
            },
        )
        recrel_assert_record_relations(
            rec2,
            expected={
                "relations": {
                    "edition": [
                        {
                            "pid_value": rec_docid_1["pid"],
                            "pid_type": "docid",
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec_docid_1["title"],
                                "edition": rec_docid_1["edition"],
                                "languages": rec_docid_1["languages"],
                                "document_type": rec_docid_1["document_type"],
                                "publication_year": rec_docid_1["publication_year"],
                            },
                        },
                        {
                            "pid_value": rec_docid_8["pid"],
                            "pid_type": "docid",
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec_docid_8["title"],
                                "edition": rec_docid_8["edition"],
                                "document_type": rec_docid_8["document_type"],
                                "publication_year": rec_docid_8["publication_year"],
                            },
                        },
                    ],
                    "other": [
                        {
                            "pid_value": first_pid_value,
                            "pid_type": first_pid_type,
                            "note": "exercise",
                            "relation_type": "other",
                            "record_metadata": {
                                "title": rec1["title"],
                                "languages": rec1["languages"],
                                "document_type": rec1["document_type"],
                                "publication_year": rec1["publication_year"],
                            },
                        }
                    ],
                }
            },
        )

    def _test_delete():
        """Test relation deletion."""
        rec1, rec2 = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "DELETE"),
            (
                first_pid_value,
                first_pid_type,
                second_pid_value,
                second_pid_type,
            ),
            payload,
        )

        rec_docid_1 = get_test_record(testdata, "documents", "docid-1")
        rec_docid_6 = get_test_record(testdata, "documents", "docid-6")
        rec_docid_8 = get_test_record(testdata, "documents", "docid-8")

        recrel_assert_record_relations(
            rec1,
            expected={
                "relations": {
                    "language": [
                        {
                            "pid_value": rec_docid_1["pid"],
                            "pid_type": "docid",
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec_docid_1["title"],
                                "edition": rec_docid_1["edition"],
                                "languages": rec_docid_1["languages"],
                                "document_type": rec_docid_1["document_type"],
                                "publication_year": rec_docid_1["publication_year"],
                            },
                        },
                        {
                            "pid_value": rec_docid_6["pid"],
                            "pid_type": "docid",
                            "relation_type": "language",
                            "record_metadata": {
                                "title": rec_docid_6["title"],
                                "document_type": rec_docid_6["document_type"],
                                "languages": rec_docid_6["languages"],
                                "publication_year": rec_docid_6["publication_year"],
                            },
                        },
                    ]
                }
            },
        )
        recrel_assert_record_relations(
            rec2,
            expected={
                "relations": {
                    "edition": [
                        {
                            "pid_value": rec_docid_1["pid"],
                            "pid_type": "docid",
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec_docid_1["title"],
                                "edition": rec_docid_1["edition"],
                                "languages": rec_docid_1["languages"],
                                "document_type": rec_docid_1["document_type"],
                                "publication_year": rec_docid_1["publication_year"],
                            },
                        },
                        {
                            "pid_value": rec_docid_8["pid"],
                            "pid_type": "docid",
                            "relation_type": "edition",
                            "record_metadata": {
                                "title": rec_docid_8["title"],
                                "edition": rec_docid_8["edition"],
                                "document_type": rec_docid_8["document_type"],
                                "publication_year": rec_docid_8["publication_year"],
                            },
                        },
                    ]
                }
            },
        )

    _test_create()
    _test_delete()
    # recreate for the next one, to have some more valuable test data
    _test_create()


def _test_sibl_invalid_relations_should_fail(
    client, json_headers, invalids, status_code=400
):
    """Test relation creation with invalid siblings should fail."""
    api_endpoint_documents = "invenio_app_ils_relations.docid_relations"
    api_endpoint_series = "invenio_app_ils_relations.serid_relations"

    for invalid in invalids:
        first_pid_value = invalid["first_pid_value"]
        first_pid_type = invalid["first_pid_type"]
        second_pid_value = invalid["second_pid_value"]
        second_pid_type = invalid["second_pid_type"]
        relation_type = invalid["relation_type"]

        api_endpoint = (
            api_endpoint_documents if first_pid_type == "docid" else api_endpoint_series
        )

        url = url_for(api_endpoint, pid_value=first_pid_value)
        payload = {
            "pid_value": second_pid_value,
            "pid_type": second_pid_type,
            "relation_type": relation_type,
        }

        res = client.post(url, headers=json_headers, data=json.dumps(payload))
        assert res.status_code == status_code
        if status_code == 400:
            error = json.loads(res.data.decode("utf-8"))
            assert "message" in error
            assert first_pid_value in error["message"]
            assert second_pid_value in error["message"]


def test_siblings_relations(client, json_headers, testdata, users):
    """Test siblings relations."""

    # only one test method to speed up tests and avoid testdata recreation at
    # each test. As drawback, testdata is not cleaned between each test, so
    # do not change the order of execution of the following tests :)

    _test_sibl_invalid_relations_should_fail(
        client,
        json_headers,
        [
            {
                "first_pid_value": "docid-1",
                "first_pid_type": "docid",
                "second_pid_value": "docid-2",
                "second_pid_type": "docid",
                "relation_type": "language",
            }
        ],
        status_code=401,
    )

    user_login(client, "librarian", users)

    # docid-1 --language--> docid-2 and docid-6
    _test_sibl_language_relation(client, json_headers)

    # docid-3 --edition--> docid-1 and docid-8
    _test_sibl_edition_relation(client, json_headers, testdata)

    # docid-2 --other--> docid-3
    _test_sibl_other_relation(client, json_headers, testdata)

    # test wrong relations
    invalids = [
        # different pid type
        {
            "first_pid_value": "docid-1",
            "first_pid_type": "docid",
            "second_pid_value": "serid-1",
            "second_pid_type": "serid",
            "relation_type": "language",
        },
        # invalid edition: document with serial
        {
            "first_pid_value": "serid-3",
            "first_pid_type": "serid",
            "second_pid_value": "docid-5",
            "second_pid_type": "docid",
            "relation_type": "edition",
        },
        # different pid type
        {
            "first_pid_value": "serid-1",
            "first_pid_type": "serid",
            "second_pid_value": "docid-1",
            "second_pid_type": "docid",
            "relation_type": "other",
        },
        # same record
        {
            "first_pid_value": "docid-6",
            "first_pid_type": "docid",
            "second_pid_value": "docid-6",
            "second_pid_type": "docid",
            "relation_type": "language",
        },
    ]
    _test_sibl_invalid_relations_should_fail(client, json_headers, invalids)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Response serializers for circulation module."""

import json

from flask import current_app


def bulk_loan_extend_responsify(serializer, mimetype):
    """Bulk loan extend serializer.

    :param schema_class: Serializer instance.
    :param mimetype: MIME type of response.
    """

    def view(extended_loans, not_extended_loans, code=200, headers=None, **kwargs):
        """Generate the response object."""
        response_data = {"extended_loans": [], "not_extended_loans": []}
        for loan in extended_loans:
            response_data["extended_loans"].append(
                serializer.serialize(loan["pid"], loan)
            )
        for loan in not_extended_loans:
            response_data["not_extended_loans"].append(
                serializer.serialize(loan["pid"], loan)
            )

        response = current_app.response_class(
            json.dumps(response_data), mimetype=mimetype
        )
        response.status_code = code

        if headers is not None:
            response.headers.extend(headers)
        return response

    return view



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test files."""

import json

from flask import url_for
from invenio_search import current_search
from six import BytesIO

from tests.helpers import user_login


def _test_response(
    client,
    req_method,
    url,
    headers,
    data,
    expected_resp_code,
    input_stream=None,
):
    """Util function testing response code."""
    kwargs = {}
    if input_stream:
        kwargs["input_stream"] = input_stream
    if data:
        res = getattr(client, req_method)(
            url, headers=headers, data=json.dumps(data), **kwargs
        )
    else:
        res = getattr(client, req_method)(url, headers=headers, **kwargs)
    assert expected_resp_code == res.status_code
    return res


def _test_data_exists(key, res):
    """Util function for testing output data."""
    data = json.loads(res.data.decode("utf-8"))["metadata"]
    assert key in data


def test_create_bucket_endpoint(client, json_headers, location, testdata, users):
    """Test GET permissions."""
    user_login(client, "admin", users)

    url_with_bucket_id = url_for(
        "invenio_app_ils_files.eitmid_bucket", pid_value="eitemid-3"
    )
    url_without_bucket_id = url_for(
        "invenio_app_ils_files.eitmid_bucket", pid_value="eitemid-4"
    )

    res1 = _test_response(client, "post", url_with_bucket_id, json_headers, None, 200)
    _test_data_exists("bucket_id", res1)
    res2 = _test_response(
        client, "post", url_without_bucket_id, json_headers, None, 201
    )
    _test_data_exists("bucket_id", res2)


def test_create_bucket_permissions(client, json_headers, location, testdata, users):
    """Test create bucket permissions."""
    url = url_for("invenio_app_ils_files.eitmid_bucket", pid_value="eitemid-1")
    _test_response(client, "post", url, json_headers, None, 401)

    test_data = [
        ("admin", "eitemid-1", 201),
        ("librarian", "eitemid-2", 201),
        ("patron1", "eitemid-2", 403),
    ]
    for user, pid, status_code in test_data:
        user_login(client, user, users)
        url = url_for("invenio_app_ils_files.eitmid_bucket", pid_value=pid)
        _test_response(client, "post", url, json_headers, None, status_code)


def test_upload_files_permissions(client, json_headers, bucket, testdata, users):
    """Test upload files permissions."""
    filename = "myfile.txt"
    data = b"hello world"
    url = url_for(
        "invenio_files_rest.object_api", bucket_id=str(bucket.id), key=filename
    )

    test_data = [
        ("anonymous", 404),
        ("admin", 200),
        ("librarian", 200),
        ("patron1", 404),
    ]
    for user, status_code in test_data:
        user_login(client, user, users)
        _test_response(
            client,
            "put",
            url,
            json_headers,
            None,
            status_code,
            input_stream=BytesIO(data),
        )


def test_download_files_permissions(client, json_headers, location, testdata, users):
    """Test download files permissions."""

    tests = [
        (
            "eitemid-4",
            [
                ("anonymous", 200),
                ("admin", 200),
                ("librarian", 200),
                ("patron1", 200),
            ],
        ),
        (
            "eitemid-5",
            [
                ("anonymous", 404),
                ("admin", 200),
                ("librarian", 200),
                ("patron1", 200),
            ],
        ),
    ]

    for pid, expected in tests:
        # Create e-item bucket
        user_login(client, "admin", users)
        url = url_for("invenio_app_ils_files.eitmid_bucket", pid_value=pid)
        res = _test_response(client, "post", url, json_headers, None, 201)
        bucket_id = json.loads(res.data)["metadata"]["bucket_id"]

        # Upload file to e-item bucket
        filename = "myfile.txt"
        data = b"hello world"
        url = url_for(
            "invenio_files_rest.object_api", bucket_id=bucket_id, key=filename
        )
        _test_response(
            client,
            "put",
            url,
            json_headers,
            None,
            200,
            input_stream=BytesIO(data),
        )

        current_search.flush_and_refresh(index="eitems")

        # Download file
        for user, status_code in expected:
            user_login(client, user, users)
            _test_response(client, "get", url, None, None, status_code)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test accessibility of resource endpoints."""

import copy
import json

from flask import url_for

from tests.helpers import user_login, validate_response

_HTTP_OK = [200, 201, 204]
ITEM_PID = "itemid-1"
ITEM_ENDPOINT = "invenio_records_rest.pitmid_item"
LIST_ENDPOINT = "invenio_records_rest.pitmid_list"


def test_items_permissions(client, testdata, item_record, json_headers, users):
    """Test items endpoints permissions."""
    dummy_item = copy.deepcopy(item_record)
    tests = [
        ("admin", _HTTP_OK, dummy_item),
        ("librarian", _HTTP_OK, dummy_item),
        ("patron1", [403], dummy_item),
        ("anonymous", [401], dummy_item),
    ]

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_create(expected_status, data):
        """Test record creation."""
        url = url_for(LIST_ENDPOINT)
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status

        if res.status_code < 400:
            record = res.get_json()["metadata"]
            return record["pid"]

    def _test_update(expected_status, data, pid):
        """Test record update."""
        pid_value = pid or ITEM_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.put(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status
        if res.status_code < 400:
            record = res.get_json()["metadata"]

    def _test_read(expected_status, pid):
        """Test record read."""
        pid_value = pid or ITEM_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_delete(expected_status, pid):
        """Test record delete."""
        pid_value = pid or ITEM_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.delete(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, expected_status, data in tests:
        user_login(client, username, users)
        _test_list(expected_status)
        pid = _test_create(expected_status, data)
        _test_update(expected_status, data, pid)
        _test_read(expected_status, pid)
        _test_delete(expected_status, pid)


def test_item_circulation(client, json_headers, testdata, users):
    """Test item circulation filtering."""

    tests = [
        ("patron1", "itemid-56", 403, False),
        ("patron2", "itemid-56", 403, True),
        ("patron1", "itemid-57", 403, True),
        ("patron2", "itemid-57", 403, False),
        ("librarian", "itemid-56", 200, False),
        ("admin", "itemid-57", 200, False),
    ]

    for username, pid, expected_resp_code, filtered in tests:
        user_login(client, username, users)
        url = url_for("invenio_records_rest.pitmid_item", pid_value=pid)
        res = validate_response(
            client, "get", url, json_headers, None, expected_resp_code
        )
        if res.status_code < 400:
            circulation = res.json["metadata"]["circulation"]
            filter_keys = ["loan_pid", "patron_pid"]
            if filtered:
                for key in filter_keys:
                    assert key not in circulation
            else:
                for key in filter_keys:
                    assert key in circulation



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Notifications messages."""
import os
import uuid

from flask import current_app
from jinja2 import TemplateError


class NotificationMsg:
    """Generate the notification message with params."""

    TEMPLATES_BASE_DIR = "invenio_app_ils/notifications"

    def __init__(self, template, ctx=None, **kwargs):
        """Build message body and HTML based on the provided template.

        The template needs to provide two blocks: title and body. An optional
        html block can also be provided for HTML rendered msgs.

        :param template: Path to the template file.
        :param ctx: A dict of params to be injected in the template.
        :param **kwargs: extra arguments.
        """
        self.template = template
        self.id = str(uuid.uuid4())

        ctx = ctx or {}
        ctx.update(
            dict(
                spa_routes=dict(
                    HOST=current_app.config["SPA_HOST"],
                    PATHS=current_app.config["SPA_PATHS"],
                )
            )
        )
        self.ctx = ctx

        tmpl = current_app.jinja_env.get_template(template)
        self.title = self.render_block(tmpl, "title").strip()
        self.body_plain = self.render_block(tmpl, "body_plain")
        try:
            self.body_html = self.render_block(tmpl, "body_html")
        except TemplateError:
            self.body_html = self.body_plain

        footer_template = current_app.config.get("ILS_NOTIFICATIONS_TEMPLATES", {}).get(
            "footer"
        )
        if footer_template:
            path = os.path.join(self.TEMPLATES_BASE_DIR, footer_template)
            footer_tmpl = current_app.jinja_env.get_template(path)
            footer_plain = self.render_block(footer_tmpl, "footer_plain")
            try:
                footer_html = self.render_block(footer_tmpl, "footer_html")
            except TemplateError:
                footer_html = footer_plain
            self.body_plain += footer_plain
            self.body_html += footer_html

        self.body_plain = self.body_plain.strip()
        self.body_html = self.body_html.strip()

    def render_block(self, template, block_name):
        """Return a Jinja2 block as a string."""
        new_context = template.new_context
        if block_name not in template.blocks:
            raise TemplateError("No block with name '{}'".format(block_name))
        lines = template.blocks[block_name](new_context(vars=self.ctx))
        return "".join(lines)

    def to_dict(self):
        """Dump obj."""
        return dict(
            id=self.id,
            title=self.title,
            body_plain=self.body_plain,
            body_html=self.body_html,
            msg_cls=self.__class__.__name__,
        )


def notification_msg_builder(template, ctx, **kwargs):
    """Factory builder to create a notification msg."""
    return NotificationMsg(template, ctx, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Circulation Loan resolvers."""

from invenio_circulation.proxies import current_circulation
from invenio_pidstore.errors import PIDDeletedError

from invenio_app_ils.circulation.utils import resolve_item_from_loan
from invenio_app_ils.documents.utils import flatten_authors
from invenio_app_ils.patrons.api import get_patron_or_unknown_dump
from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.records.jsonresolvers.api import (
    get_field_value_for_record as get_field_value,
)
from invenio_app_ils.records.jsonresolvers.api import get_pid_or_default, pick


def item_resolver(loan_pid):
    """Resolve an Item given a Loan PID."""
    Loan = current_circulation.loan_record_cls
    loan = Loan.get_record_by_pid(loan_pid)
    if not loan.get("item_pid"):
        return {}

    try:
        # can resolve to an Item or BorrowingRequest
        item = resolve_item_from_loan(loan["item_pid"])
    except PIDDeletedError:
        item = {}
    else:
        # if it the item is a BorrowingRequest, then some of these
        # fields might not be there
        item = pick(
            item,
            "barcode",
            "description",
            "document_pid",
            "medium",
            "pid",
        )

    return item


@get_pid_or_default(default_value=dict())
def loan_patron_resolver(loan_pid):
    """Resolve a Patron given a Loan PID."""
    Loan = current_circulation.loan_record_cls
    try:
        patron_pid = get_field_value(Loan, loan_pid, "patron_pid")
    except KeyError:
        return {}

    return get_patron_or_unknown_dump(patron_pid)


@get_pid_or_default(default_value=dict())
def document_resolver(loan_pid):
    """Resolve a Document given a Loan PID."""
    Loan = current_circulation.loan_record_cls
    try:
        document_pid = get_field_value(Loan, loan_pid, "document_pid")
    except KeyError:
        return {}

    Document = current_app_ils.document_record_cls
    try:
        document = Document.get_record_by_pid(document_pid)
    except PIDDeletedError:
        obj = {}
    else:
        obj = pick(
            document,
            "authors",
            "cover_metadata",
            "document_type",
            "edition",
            "identifiers",
            "open_access",
            "pid",
            "publication_year",
            "title",
        )
        obj["authors"] = flatten_authors(obj["authors"])

    return obj



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test metadata extensions."""

import pytest
from invenio_indexer.api import RecordIndexer
from invenio_records_rest.schemas.fields import DateString, SanitizedUnicode
from marshmallow import ValidationError
from marshmallow.fields import Bool, Integer, List

from invenio_app_ils.documents.api import Document
from invenio_app_ils.records.metadata_extensions import MetadataExtensions


@pytest.fixture(scope="module")
def app_config(app_config):
    """Override conftest.py's app_config"""
    app_config["ILS_RECORDS_METADATA_NAMESPACES"] = {
        "document": {
            "accelerator_experiments": {
                "@context": "https://example.com/accelerator_experiments/terms"
            },
            "standard_status": {
                "@context": "https://example.com/standard_status/terms"
            },
        }
    }

    app_config["ILS_RECORDS_METADATA_EXTENSIONS"] = {
        "document": {
            "schema": Document._schema,
            "fields": {
                "accelerator_experiments_accelerator": {
                    "elasticsearch": "keyword",
                    "marshmallow": SanitizedUnicode(required=True),
                },
                "accelerator_experiments_project": {
                    "elasticsearch": "text",
                    "marshmallow": SanitizedUnicode(),
                },
                "accelerator_experiments_number_in_sequence": {
                    "elasticsearch": "long",
                    "marshmallow": Integer(),
                },
                "accelerator_experiments_scientific_sequence": {
                    "elasticsearch": "long",
                    "marshmallow": List(Integer()),
                },
                "standard_status_original_presentation_date": {
                    "elasticsearch": "date",
                    "marshmallow": DateString(),
                },
                "standard_status_right_or_wrong": {
                    "elasticsearch": "boolean",
                    "marshmallow": Bool(),
                },
            },
        }
    }

    return app_config


def assert_unordered_equality(array_dict1, array_dict2):
    array1 = sorted(array_dict1, key=lambda d: d["key"])
    array2 = sorted(array_dict2, key=lambda d: d["key"])
    assert array1 == array2


def test_metadata_extensions(db, es, testdata):
    """Tests that a Record is indexed into Elasticsearch properly.
    - Tests that the before_record_index_hook is registered properly.
    - Tests add_es_metadata_extensions.
    - Tests jsonschema validates correctly
    - Tests that retrieved record document is fine.
    NOTE:
        - es fixture depends on appctx fixture, so we are in app context
        - this test requires a running ES instance
    """
    document = testdata["documents"][0]
    data = {
        "extensions": {
            "accelerator_experiments_accelerator": "LHCb",
            "accelerator_experiments_project": "A project for experiment.",
            "accelerator_experiments_number_in_sequence": 3,
            "accelerator_experiments_scientific_sequence": [1, 1, 2, 3, 5, 8],
            "standard_status_original_presentation_date": "2019-02-14",
            "standard_status_right_or_wrong": True,
        }
    }
    document.update(data)
    document.validate()
    db.session.commit()
    indexer = RecordIndexer()
    index_result = indexer.index(document)

    _index = index_result["_index"]
    _id = index_result["_id"]
    es_doc = es.get(index=_index, id=_id)
    source = es_doc["_source"]
    expected_keywords = [
        {"key": "accelerator_experiments_accelerator", "value": "LHCb"},
    ]
    expected_texts = [
        {
            "key": "accelerator_experiments_project",
            "value": "A project for experiment.",
        },
    ]
    expected_longs = [
        {"key": "accelerator_experiments_number_in_sequence", "value": 3},
        {
            "key": "accelerator_experiments_scientific_sequence",
            "value": [1, 1, 2, 3, 5, 8],
        },
    ]
    expected_dates = [
        {
            "key": "standard_status_original_presentation_date",
            "value": "2019-02-14",
        },
    ]
    expected_booleans = [
        {"key": "standard_status_right_or_wrong", "value": True},
    ]
    assert_unordered_equality(source["extensions_keywords"], expected_keywords)
    assert_unordered_equality(source["extensions_texts"], expected_texts)
    assert_unordered_equality(source["extensions_longs"], expected_longs)
    assert_unordered_equality(source["extensions_dates"], expected_dates)
    assert_unordered_equality(source["extensions_booleans"], expected_booleans)


def test_extensions():
    """Test metadata extensions schema."""
    ILS_RECORDS_METADATA_NAMESPACES = {
        "accelerator_experiments": {
            "@context": "https://example.com/accelerator_experiments/terms"
        },
        "standard_status": {"@context": "https://example.com/standard_status/terms"},
    }

    ILS_RECORDS_METADATA_EXTENSIONS = {
        "accelerator_experiments_accelerator": {
            "elasticsearch": "keyword",
            "marshmallow": SanitizedUnicode(required=True),
        },
        "accelerator_experiments_project": {
            "elasticsearch": "text",
            "marshmallow": SanitizedUnicode(),
        },
        "accelerator_experiments_number_in_sequence": {
            "elasticsearch": "long",
            "marshmallow": Integer(),
        },
        "accelerator_experiments_scientific_sequence": {
            "elasticsearch": "long",
            "marshmallow": List(Integer()),
        },
        "standard_status_original_presentation_date": {
            "elasticsearch": "date",
            "marshmallow": DateString(),
        },
        "standard_status_right_or_wrong": {
            "elasticsearch": "boolean",
            "marshmallow": Bool(),
        },
    }

    extensions = MetadataExtensions(
        ILS_RECORDS_METADATA_NAMESPACES, ILS_RECORDS_METADATA_EXTENSIONS
    )
    ExtensionsSchema = extensions.to_schema()

    # Minimal if not absent
    valid_minimal = {"accelerator_experiments_accelerator": "LHCb"}

    data = ExtensionsSchema().load(valid_minimal)

    assert data == valid_minimal

    # Full
    valid_full = {
        "accelerator_experiments_accelerator": "LHCb",
        "accelerator_experiments_project": "A project for experiment.",
        "accelerator_experiments_number_in_sequence": 3,
        "accelerator_experiments_scientific_sequence": [1, 1, 2, 3, 5, 8],
        "standard_status_original_presentation_date": "2019-02-14",
        "standard_status_right_or_wrong": True,
    }

    data = ExtensionsSchema().load(valid_full)

    assert data == valid_full

    # Invalid
    invalid_number_in_sequence = {
        "accelerator_experiments_accelerator": "LHCb",
        "accelerator_experiments_scientific_sequence": [1, "l", 2, 3, 5, 8],
    }
    with pytest.raises(ValidationError):
        data = ExtensionsSchema().load(invalid_number_in_sequence)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Resolve the referenced license."""

from urllib.parse import unquote_plus

import jsonresolver
from werkzeug.routing import Rule

from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.vocabularies.api import VOCABULARY_TYPE_LICENSE

# Note: there must be only one resolver per file,
# otherwise only the last one is registered

REF_LICENSE_URL_ABSOLUTE_PATH = "{scheme}://{host}/api/resolver/licenses/{license_id}"
REF_LICENSE_URL_RELATIVE_PATH = "/api/resolver/licenses/<license_id>"


@jsonresolver.hookimpl
def jsonresolver_loader(url_map):
    """Resolve the referred license for a record."""
    from flask import current_app

    def license_resolver(license_id):
        """Search and return the license info."""
        _id = unquote_plus(license_id)
        vocabulary_search = current_app_ils.vocabulary_search_cls()

        search = vocabulary_search.search_by_type_and_key(
            type=VOCABULARY_TYPE_LICENSE, key=unquote_plus(_id)
        )
        search_result = search.execute()
        total = search_result.hits.total.value

        _license = dict()
        if total == 1:
            hit = search_result.hits[0].to_dict()
            _license = dict(
                id=_id,
                maintainer=hit["data"].get("maintainer", ""),
                status=hit["data"].get("status", ""),
                title=hit["text"],
                url=hit["data"].get("url", ""),
            )
        else:
            _license = dict(
                id=_id,
                maintainer="",
                status="",
                title="",
                url="",
            )

        return _license

    url_map.add(
        Rule(
            REF_LICENSE_URL_RELATIVE_PATH,
            endpoint=license_resolver,
            host=current_app.config.get("JSONSCHEMAS_HOST"),
        )
    )



from invenio_app_ils.document_requests.notifications.api import (
    send_document_request_notification,
)


def test_notification_document_request_filter(app_with_notifs, mocker, users, testdata):
    """Test notification filtering."""

    def filter_document_request_messages(record, action, **kwargs):
        """Define notification filter."""
        if action == "decline":
            return False
        return True

    send_mocked = mocker.patch("invenio_app_ils.notifications.backends.mail.send")
    send_mocked.__name__ = "send"
    send_mocked.__annotations__ = "send"
    backends = mocker.patch(
        "invenio_app_ils.notifications.api._get_notification_backends",
        return_value=[send_mocked],
    )
    document_request = testdata["document_requests"][0]
    with app_with_notifs.app_context():
        # remove footer
        app_with_notifs.config["ILS_NOTIFICATIONS_TEMPLATES"] = {}
        # set filtering on the document requests
        app_with_notifs.config["ILS_NOTIFICATIONS_FILTER_DOCUMENT_REQUEST"] = (
            filter_document_request_messages
        )

        send_document_request_notification(
            document_request, action="decline", msg_extra_ctx={}
        )

        assert not send_mocked.apply_async.called



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test accesibility of resource endpoints."""

import json

from flask import url_for

from tests.helpers import user_login, validate_response

NEW_INTERNAL_LOCATION = {
    "location_pid": "locid-1",
    "legacy_ids": ["Test legacy id"],
    "name": "My test internal location",
    "physical_location": "Left from the right building",
    "notes": "In house",
}


def test_post_internal_location(client, json_headers, testdata, users):
    """Test POST of internal_location."""
    user_login(client, "admin", users)
    url = url_for("invenio_records_rest.ilocid_list")
    res = validate_response(
        client, "post", url, json_headers, NEW_INTERNAL_LOCATION, 201
    )
    data = json.loads(res.data.decode("utf-8"))["metadata"]
    assert "name" in data["location"]


def test_post_partial_internal_location(client, json_headers, testdata, users):
    """Test POST of internal_location without all required data."""
    user_login(client, "admin", users)
    del NEW_INTERNAL_LOCATION["location_pid"]
    url = url_for("invenio_records_rest.ilocid_list")
    validate_response(client, "post", url, json_headers, NEW_INTERNAL_LOCATION, 400)


def test_post_item(client, json_headers, testdata, users, item_record):
    """Test POST of an item."""
    user_login(client, "admin", users)
    url = url_for("invenio_records_rest.pitmid_list")
    del item_record["pid"]
    res = validate_response(client, "post", url, json_headers, item_record, 201)
    data = json.loads(res.data.decode("utf-8"))["metadata"]
    assert "name" in data["internal_location"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test bulk loan extension."""

from datetime import timedelta

import arrow
from flask import current_app
from invenio_circulation.api import Loan
from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from invenio_circulation.proxies import current_circulation
from invenio_search import current_search

from invenio_app_ils.circulation.api import bulk_extend_loans
from invenio_app_ils.items.api import ITEM_PID_TYPE, Item
from invenio_app_ils.proxies import current_app_ils
from tests.api.conftest import _create_records
from tests.helpers import user_login

pid_start_value = 200
amount_of_ongoing_loans = 3
amount_of_overdue_loans = 3
pid_end_value = pid_start_value + amount_of_ongoing_loans + amount_of_overdue_loans


def create_loans(start=pid_start_value, end=pid_end_value):
    ongoing_loans = []
    overdue_loans = []

    day_range_loan_extendable = current_app.config[
        "ILS_CIRCULATION_LOAN_WILL_EXPIRE_DAYS"
    ]

    now = arrow.utcnow()

    start_date_expiring = now.isoformat()
    start_date_overdue = (now - timedelta(days=5)).isoformat()

    end_date_expiring = (
        now + timedelta(days=day_range_loan_extendable - 1)
    ).isoformat()
    end_date_overdue = (now - timedelta(days=2)).isoformat()

    def loan_generator(pid_value, overdue_loan=False):
        end_date = end_date_overdue if overdue_loan else end_date_expiring
        start_date = start_date_overdue if overdue_loan else start_date_expiring

        loan = {
            "document_pid": "docid-1",
            "end_date": end_date,
            "item_pid": {"type": "pitmid", "value": f"itemid-{pid_value}"},
            "patron_pid": "1",
            "pickup_location_pid": "locid-1",
            "pid": f"loanid-{pid_value}",
            "start_date": start_date,
            "state": "ITEM_ON_LOAN",
            "transaction_location_pid": "locid-1",
            "transaction_user_pid": "1",
            "patron": {"email": "patron1@test.ch"},
        }

        return loan

    for number in range(start, end):
        if len(ongoing_loans) < amount_of_ongoing_loans:
            ongoing_loans.append(loan_generator(number))
        elif len(overdue_loans) < amount_of_overdue_loans:
            overdue_loans.append(loan_generator(number, True))

    return ongoing_loans + overdue_loans


def create_items(start=pid_start_value, end=pid_end_value):
    items = []

    def items_generator(pid_value):
        item = {
            "pid": f"itemid-{pid_value}",
            "created_by": {"type": "script", "value": "demo"},
            "barcode": f"123456{pid_value}89-1",
            "document_pid": "docid-1",
            "internal_location_pid": "ilocid-1",
            "circulation_restriction": "NO_RESTRICTION",
            "medium": "NOT_SPECIFIED",
            "status": "CAN_CIRCULATE",
            "document": {},
        }
        return item

    for number in range(start, end):
        items.append(items_generator(number))

    return items


def _prepare_data(db, repeated_run_number=0):
    total_loans = amount_of_ongoing_loans + amount_of_overdue_loans
    added_value = repeated_run_number * total_loans + repeated_run_number

    new_pid_start_value = pid_start_value + added_value
    new_pid_end_value = pid_end_value + added_value

    loans = create_loans(new_pid_start_value, new_pid_end_value)
    items = create_items(new_pid_start_value, new_pid_end_value)

    bulk_index_items = _create_records(db, items, Item, ITEM_PID_TYPE)
    bulk_index_loans = _create_records(db, loans, Loan, CIRCULATION_LOAN_PID_TYPE)

    for item in bulk_index_items:
        current_app_ils.item_indexer.index(item)

    for loan in bulk_index_loans:
        current_circulation.loan_indexer().index(loan)

    current_search.flush_and_refresh(index="*")

    return {"loans": bulk_index_loans, "items": bulk_index_items}


def test_loan_extend_permissions(
    client, json_headers, users, testdata, loan_params, db
):
    def _run_asserts(extended_loans, loans, patron_pid="1"):
        assert len(extended_loans) >= len(loans)

        for extended_loan in extended_loans:
            for loan in loans:
                if loan["pid"] == extended_loan["pid"]:
                    assert extended_loan["state"] == "ITEM_ON_LOAN"
                    assert extended_loan["patron_pid"] == str(patron_pid)
                    assert extended_loan["end_date"] > loan["end_date"]

    def test_patron1_can_extend_own_loans():
        patron_1_ongoing_loans = _prepare_data(db)["loans"]

        user_login(client, "patron1", users)

        extended_loans, _ = bulk_extend_loans(patron_pid=1)

        _run_asserts(extended_loans, patron_1_ongoing_loans)

    def test_library_can_extend_any_loan():
        patron_1_ongoing_loans = _prepare_data(db, repeated_run_number=1)["loans"]

        user_login(client, "librarian", users)

        extended_loans, _ = bulk_extend_loans(patron_pid=1)

        _run_asserts(extended_loans, patron_1_ongoing_loans)

    def test_patron2_cannot_extend_patron1_loans():
        _prepare_data(db, repeated_run_number=2)

        user_login(client, "patron2", users)

        patron1_pid = "1"

        extended_loans, _ = bulk_extend_loans(patron1_pid)

        assert len(extended_loans) == 0

    test_patron1_can_extend_own_loans()
    test_library_can_extend_any_loan()
    test_patron2_cannot_extend_patron1_loans()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS Document indexer APIs."""

from datetime import datetime

from celery import shared_task
from flask import current_app
from invenio_circulation.pidstore.pids import CIRCULATION_LOAN_PID_TYPE
from invenio_circulation.proxies import current_circulation
from invenio_circulation.search.api import search_by_pid as search_loans_by_pid
from invenio_indexer.api import RecordIndexer
from invenio_pidstore.errors import PIDDeletedError

from invenio_app_ils.acquisition.api import ORDER_PID_TYPE
from invenio_app_ils.acquisition.proxies import current_ils_acq
from invenio_app_ils.document_requests.api import DOCUMENT_REQUEST_PID_TYPE
from invenio_app_ils.documents.api import DOCUMENT_PID_TYPE
from invenio_app_ils.eitems.api import EITEM_PID_TYPE
from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE
from invenio_app_ils.ill.proxies import current_ils_ill
from invenio_app_ils.indexer import ReferencedRecordsIndexer
from invenio_app_ils.items.api import ITEM_PID_TYPE
from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.records.api import IlsRecord


def get_items(document_pid):
    """Get referenced items."""
    referenced = []
    item_search_cls = current_app_ils.item_search_cls
    item_record_cls = current_app_ils.item_record_cls
    for item in (
        item_search_cls().search_by_document_pid(document_pid=document_pid).scan()
    ):
        item = item_record_cls.get_record_by_pid(item["pid"])
        referenced.append(dict(pid_type=ITEM_PID_TYPE, record=item))
    return referenced


def get_eitems(document_pid):
    """Get referenced items."""
    referenced = []
    eitem_search_cls = current_app_ils.eitem_search_cls
    eitem_record_cls = current_app_ils.eitem_record_cls
    for eitem in (
        eitem_search_cls().search_by_document_pid(document_pid=document_pid).scan()
    ):
        eitem = eitem_record_cls.get_record_by_pid(eitem["pid"])
        referenced.append(dict(pid_type=EITEM_PID_TYPE, record=eitem))
    return referenced


def get_related_records(document_pid):
    """Get referenced records via relations."""
    referenced = []
    doc_record_cls = current_app_ils.document_record_cls
    try:
        record = doc_record_cls.get_record_by_pid(document_pid)
    except PIDDeletedError:
        # Document might have been deleted while reindexing asynchronously.
        return referenced
    relations = record.relations
    for _, related_records in relations.items():
        for obj in related_records:
            rec = IlsRecord.get_record_by_pid(
                obj["pid_value"], pid_type=obj["pid_type"]
            )
            referenced.append(dict(pid_type=obj["pid_type"], record=rec))
    return referenced


def get_document_requests(document_pid):
    """Get referenced documents requests."""
    referenced = []
    docreq_search_cls = current_app_ils.document_request_search_cls
    docreq_record_cls = current_app_ils.document_request_record_cls
    for request in (
        docreq_search_cls().search_by_document_pid(document_pid=document_pid).scan()
    ):
        docreq = docreq_record_cls.get_record_by_pid(request["pid"])
        referenced.append(dict(pid_type=DOCUMENT_REQUEST_PID_TYPE, record=docreq))
    return referenced


def get_loans(document_pid):
    """Get referenced loans."""
    referenced = []
    loan_record_cls = current_circulation.loan_record_cls
    for loan in search_loans_by_pid(document_pid=document_pid).scan():
        loan = loan_record_cls.get_record_by_pid(loan["pid"])
        referenced.append(dict(pid_type=CIRCULATION_LOAN_PID_TYPE, record=loan))
    return referenced


def get_acquisition_orders(document_pid):
    """Get referenced acquisition orders."""
    referenced = []
    order_record_cls = current_ils_acq.order_record_cls
    order_search_cls = current_ils_acq.order_search_cls
    search = order_search_cls().search_by_document_pid(document_pid)
    for hit in search.scan():
        order = order_record_cls.get_record_by_pid(hit["pid"])
        referenced.append(dict(pid_type=ORDER_PID_TYPE, record=order))
    return referenced


def get_ill_borrowing_requests(document_pid):
    """Get referenced ILL borrowing requests."""
    referenced = []
    brw_req_record_cls = current_ils_ill.borrowing_request_record_cls
    brw_req_search_cls = current_ils_ill.borrowing_request_search_cls
    search = brw_req_search_cls().search_by_document_pid(document_pid)
    for hit in search.scan():
        brw_req = brw_req_record_cls.get_record_by_pid(hit["pid"])
        referenced.append(dict(pid_type=BORROWING_REQUEST_PID_TYPE, record=brw_req))
    return referenced


@shared_task(ignore_result=True)
def index_referenced_records(document):
    """Index referenced records."""
    indexer = ReferencedRecordsIndexer()

    document_pid = document["pid"]
    indexed = dict(pid_type=DOCUMENT_PID_TYPE, record=document)

    # keep loans and items as first
    # Note: this can be quite inefficient because it potentially retrieves
    # a lot of records. Might be a place where performance can be improved
    # (bulk indexing, less DB queries, etc...)
    indexer.index(indexed, get_loans(document_pid))
    indexer.index(indexed, get_items(document_pid))
    indexer.index(indexed, get_document_requests(document_pid))
    indexer.index(indexed, get_eitems(document_pid))
    indexer.index(indexed, get_related_records(document_pid))
    indexer.index(indexed, get_acquisition_orders(document_pid))
    indexer.index(indexed, get_ill_borrowing_requests(document_pid))


class DocumentIndexer(RecordIndexer):
    """Indexer class for Document record."""

    def index(self, document, arguments=None, **kwargs):
        """Index a Document."""
        super().index(document)
        eta = datetime.utcnow() + current_app.config["ILS_INDEXER_TASK_DELAY"]
        index_referenced_records.apply_async((document,), eta=eta)


@shared_task(ignore_result=True)
def index_document(document_pid):
    """Async document indexer."""
    document_class = current_app_ils.document_record_cls

    document_indexer = current_app_ils.document_indexer
    document = document_class.get_record_by_pid(document_pid)
    document_indexer.index(document)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-App-Ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Acquisition exceptions."""

from invenio_app_ils.errors import IlsException


class AcquisitionError(IlsException):
    """Raised when there is an error with Acquisition."""

    def __init__(self, description):
        """Initialize exception."""
        super().__init__(description=description)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test ILS APIs."""

import pytest
from flask import url_for
from invenio_accounts.models import User

from invenio_app_ils.documents.api import Document
from invenio_app_ils.patrons.api import patron_exists
from invenio_app_ils.proxies import current_app_ils
from invenio_app_ils.records.api import IlsRecord
from invenio_app_ils.series.api import Series
from tests.helpers import get_test_record


def test_apis(client, json_headers, testdata):
    """Test various APIs."""

    def test_patron_exists():
        """Test return True if item exists."""
        test_patron = User.query.all()[0]
        assert patron_exists(int(test_patron.id))
        assert patron_exists(str(test_patron.id))

        with pytest.raises(AssertionError):
            assert not patron_exists("not a number")
            assert not patron_exists("-1")
            assert not patron_exists("0")

    def test_get_record_by_pid():
        """Test get_record_by_pid."""
        tests = [("docid-1", Document), ("serid-1", Series)]

        for pid, cls in tests:
            record = cls.get_record_by_pid(pid)

            assert isinstance(record, cls)
            assert record["pid"] == pid
            assert record._pid_type == cls._pid_type

    def test_get_record_by_pid_and_pid_type():
        """Test get_record_by_pid with pid_type."""
        tests = [("docid-1", "docid", Document), ("serid-1", "serid", Series)]

        for pid, pid_type, expected_cls in tests:
            record = IlsRecord.get_record_by_pid(pid, pid_type=pid_type)

            assert isinstance(record, expected_cls)
            assert record["pid"] == pid
            assert record._pid_type == pid_type

    def test_get_default_location_pid():
        """Asset that the default location is the first created."""
        first = get_test_record(testdata, "locations", "locid-1")
        pid_value, _ = current_app_ils.get_default_location_pid
        assert pid_value == first["pid"]

    def test_no_etag():
        """Assert that the response headers do not contain ETag."""
        url = url_for("invenio_records_rest.docid_item", pid_value="docid-1")
        res = client.get(url, headers=json_headers)
        assert res.cache_control.no_cache
        assert res.get_etag() == (None, None)
        assert "Last-Modified" not in res.headers

    test_patron_exists()
    test_get_record_by_pid()
    test_get_record_by_pid_and_pid_type()
    test_get_default_location_pid()
    test_no_etag()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""DocumentRequest indexer APIs."""

from datetime import datetime

from celery import shared_task
from flask import current_app
from invenio_indexer.api import RecordIndexer

from invenio_app_ils.documents.api import DOCUMENT_PID_TYPE
from invenio_app_ils.indexer import ReferencedRecordsIndexer
from invenio_app_ils.proxies import current_app_ils

from .api import DOCUMENT_REQUEST_PID_TYPE


@shared_task(ignore_result=True)
def index_referenced_records(docreq):
    """Index referenced records."""
    indexer = ReferencedRecordsIndexer()
    indexed = dict(pid_type=DOCUMENT_REQUEST_PID_TYPE, record=docreq)

    # fetch and index the document
    document_pid = docreq.get("document_pid")
    referenced = []
    if document_pid:
        document_cls = current_app_ils.document_record_cls
        document = document_cls.get_record_by_pid(document_pid)
        referenced.append(dict(pid_type=DOCUMENT_PID_TYPE, record=document))

    indexer.index(indexed, referenced)


class DocumentRequestIndexer(RecordIndexer):
    """Indexer class for DocumentRequest record."""

    def index(self, docreq, arguments=None, **kwargs):
        """Index a DocumentRequest."""
        super().index(docreq)
        eta = datetime.utcnow() + current_app.config["ILS_INDEXER_TASK_DELAY"]
        index_referenced_records.apply_async((docreq,), eta=eta)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Circulation receivers."""

from invenio_circulation.proxies import current_circulation
from invenio_circulation.signals import loan_replace_item, loan_state_changed

from invenio_app_ils.circulation.notifications.api import send_loan_notification
from invenio_app_ils.circulation.utils import resolve_item_from_loan
from invenio_app_ils.ill.api import BORROWING_REQUEST_PID_TYPE
from invenio_app_ils.ill.proxies import current_ils_ill
from invenio_app_ils.indexer import wait_es_refresh
from invenio_app_ils.items.api import ITEM_PID_TYPE
from invenio_app_ils.proxies import current_app_ils


def register_circulation_signals():
    """Register Circulation signal."""
    loan_state_changed.connect(send_notification_after_loan_change, weak=False)
    loan_replace_item.connect(index_after_loan_replace_item, weak=False)


def index_after_loan_replace_item(_, old_item_pid, new_item_pid):
    """Register Circulation signal to index item."""
    loan_index = current_circulation.loan_search_cls.Meta.index
    # When indexing items, the `circulation` field will perform a search on
    # loans index. Wait for the loan index to be refreshed.
    wait_es_refresh(loan_index)

    if old_item_pid:
        rec = resolve_item_from_loan(old_item_pid)
        indexer = None
        if old_item_pid["type"] == ITEM_PID_TYPE:
            indexer = current_app_ils.item_indexer
        elif old_item_pid["type"] == BORROWING_REQUEST_PID_TYPE:
            indexer = current_ils_ill.borrowing_request_indexer_cls()
        indexer.index(rec)

    if new_item_pid:
        rec = resolve_item_from_loan(new_item_pid)
        indexer = None
        if new_item_pid["type"] == ITEM_PID_TYPE:
            indexer = current_app_ils.item_indexer
        elif new_item_pid["type"] == BORROWING_REQUEST_PID_TYPE:
            indexer = current_ils_ill.borrowing_request_indexer_cls()
        indexer.index(rec)


def send_notification_after_loan_change(_, initial_loan, loan, trigger, **kwargs):
    """Send email notification when the loan changes."""
    send_notification = kwargs.get("send_notification", True)
    if send_notification:
        send_loan_notification(
            action=trigger,
            loan=loan,
            msg_extra_ctx=dict(initial_loan=initial_loan),
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Circulation is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Tests eitems CRUD."""

import pytest

from invenio_app_ils.eitems.api import EItem
from invenio_app_ils.errors import DocumentNotFoundError


def test_eitem_refs(app, testdata):
    """Test creation of an eitem."""
    eitem = EItem.create(
        dict(
            pid="eitemid-99",
            eitem_type="E-BOOK",
            document_pid="docid-1",
            created_by=dict(type="script", value="demo"),
        )
    )
    assert "$schema" in eitem
    assert "document" in eitem and "$ref" in eitem["document"]

    eitem = EItem.get_record_by_pid("eitemid-4")
    eitem = eitem.replace_refs()
    assert "document" in eitem and eitem["document"]["title"]


def test_eitem_validation(db, testdata):
    """Test validation when updating an eitem."""
    eitem_pid = testdata["eitems"][0]["pid"]
    eitem = EItem.get_record_by_pid(eitem_pid)

    # change document pid
    eitem["document_pid"] = "not_found_doc"
    with pytest.raises(DocumentNotFoundError):
        eitem.commit()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS ILL views."""

from flask import Blueprint, current_app
from invenio_db import db
from invenio_records_rest.utils import obj_or_import_string
from invenio_records_rest.views import need_record_permission, pass_record
from invenio_rest import ContentNegotiatedMethodView

from invenio_app_ils.permissions import need_permissions

from .api import BORROWING_REQUEST_PID_TYPE
from .errors import ILLError
from .loaders import (
    patron_loan_create_action_loader,
    patron_loan_extension_accept_loader,
    patron_loan_extension_decline_loader,
    patron_loan_extension_request_loader,
)
from .notifications.api import send_ill_notification
from .proxies import current_ils_ill


def create_ill_blueprint(app):
    """Create ILL blueprint."""
    blueprint = Blueprint(
        "invenio_app_ils_ill",
        __name__,
        template_folder="templates",
        url_prefix="",
    )

    endpoints = app.config.get("RECORDS_REST_ENDPOINTS", [])
    options = endpoints.get(BORROWING_REQUEST_PID_TYPE, {})
    default_media_type = options.get("default_media_type", "")
    rec_serializers = options.get("record_serializers", {})
    serializers = {
        mime: obj_or_import_string(func) for mime, func in rec_serializers.items()
    }

    def create_view(view_cls, path, loader):
        """Create view for an action."""
        view = view_cls.as_view(
            view_cls.view_name.format(BORROWING_REQUEST_PID_TYPE),
            serializers=serializers,
            default_media_type=default_media_type,
            ctx=dict(loader=loader),
        )
        blueprint.add_url_rule(
            "{0}/{1}".format(options["item_route"], path),
            view_func=view,
            methods=["POST"],
        )

    create_view(
        PatronLoanCreateActionResource,
        "patron-loan/create",
        patron_loan_create_action_loader,
    )
    create_view(
        RequestPatronLoanExtensionResource,
        "patron-loan/extension/request",
        patron_loan_extension_request_loader,
    )
    create_view(
        AcceptPatronLoanExtensionResource,
        "patron-loan/extension/accept",
        patron_loan_extension_accept_loader,
    )
    create_view(
        DeclinePatronLoanExtensionResource,
        "patron-loan/extension/decline",
        patron_loan_extension_decline_loader,
    )

    return blueprint


class ILLActionResource(ContentNegotiatedMethodView):
    """ILL action resource."""

    def __init__(self, serializers, ctx, *args, **kwargs):
        """Constructor."""
        super().__init__(serializers, *args, **kwargs)
        for key, value in ctx.items():
            setattr(self, key, value)


class PatronLoanCreateActionResource(ILLActionResource):
    """Create loan action resource."""

    view_name = "{}_create_loan"

    @pass_record
    @need_permissions("ill-brwreq-patron-loan-create")
    def post(self, pid, record, **kwargs):
        """Create loan action implementation."""
        data = self.loader()

        record.patron_loan.create(
            start_date=data["loan_start_date"],
            end_date=data["loan_end_date"],
            transaction_location_pid=data["transaction_location_pid"],
        )

        record.commit()
        db.session.commit()
        current_ils_ill.borrowing_request_indexer_cls().index(record)
        return self.make_response(pid, record, 200)


class ILLPatronLoanExtensionActionResource(ILLActionResource):
    """ILL extension action resource."""

    def validate_loan(self, record):
        """Validate that the extension action can be performed."""
        loan = record.patron_loan.get()
        is_loan_active = (
            loan["state"] in current_app.config["CIRCULATION_STATES_LOAN_ACTIVE"]
        )
        if not is_loan_active:
            raise ILLError("This interlibrary loan is not active.")

        return loan


class RequestPatronLoanExtensionResource(ILLPatronLoanExtensionActionResource):
    """Request extensions endpoint."""

    view_name = "{}_request_extension"

    def request_extension_permission_factory(self, record):
        """Request extension permission factory."""
        action = "ill-brwreq-patron-loan-extension-request"
        permissions = current_app.config["ILS_VIEWS_PERMISSIONS_FACTORY"]
        view_permission = permissions(action)
        return view_permission(record)

    @pass_record
    @need_record_permission("request_extension_permission_factory")
    def post(self, pid, record, **kwargs):
        """Request extension action implementation."""
        self.loader()
        self.validate_loan(record)

        record.patron_loan.extension.request()

        record.commit()
        db.session.commit()
        current_ils_ill.borrowing_request_indexer_cls().index(record)
        send_ill_notification(record, action="extension_requested")
        return self.make_response(pid, record, 200)


class AcceptPatronLoanExtensionResource(ILLPatronLoanExtensionActionResource):
    """Accept extensions endpoint."""

    view_name = "{}_accept_extension"

    @pass_record
    @need_permissions("ill-brwreq-patron-loan-extension-accept")
    def post(self, pid, record, **kwargs):
        """Accept extension action implementation."""
        data = self.loader()
        loan = self.validate_loan(record)

        record.patron_loan.extension.accept(
            data["loan_end_date"],
            transaction_location_pid=data["transaction_location_pid"],
            loan=loan,
        )

        record.commit()
        db.session.commit()
        current_ils_ill.borrowing_request_indexer_cls().index(record)
        send_ill_notification(record, action="extension_accepted")
        return self.make_response(pid, record, 200)


class DeclinePatronLoanExtensionResource(ILLPatronLoanExtensionActionResource):
    """Decline extensions endpoint."""

    view_name = "{}_decline_extension"

    @pass_record
    @need_permissions("ill-brwreq-patron-loan-extension-decline")
    def post(self, pid, record, **kwargs):
        """Decline extension action implementation."""
        self.loader()
        self.validate_loan(record)

        record.patron_loan.extension.decline()

        record.commit()
        db.session.commit()
        current_ils_ill.borrowing_request_indexer_cls().index(record)
        send_ill_notification(record, action="extension_declined")
        return self.make_response(pid, record, 200)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test records relations parent child."""

import json

import pytest
from flask import url_for

from invenio_app_ils.errors import RecordRelationsError
from invenio_app_ils.records_relations.api import RecordRelationsParentChild
from invenio_app_ils.relations.api import Relation
from tests.helpers import user_login

from .helpers import (
    recrel_assert_record_relations,
    recrel_choose_endpoints_and_do_request,
)


def _test_pc_mm_document(client, json_headers):
    """Test create and delete relations MM->Document."""

    parent_pid_value = "serid-1"
    parent_pid_type = "serid"
    child_pid_value = "docid-1"
    child_pid_type = "docid"
    relation_type = "multipart_monograph"

    payload = {
        "parent_pid_value": parent_pid_value,
        "parent_pid_type": parent_pid_type,
        "child_pid_value": child_pid_value,
        "child_pid_type": child_pid_type,
        "relation_type": relation_type,
    }

    def _test_create_mm_document(create_using_pid1=True):
        """Test relation creation of Multipart Monograph and Document."""

        parent, child = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                parent_pid_value,
                parent_pid_type,
                child_pid_value,
                child_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(parent, expected={"relations": {}})

        recrel_assert_record_relations(
            child,
            expected={
                "relations": {
                    "multipart_monograph": [
                        {
                            "pid_value": parent_pid_value,
                            "pid_type": parent_pid_type,
                            "relation_type": "multipart_monograph",
                            "record_metadata": {
                                "title": parent["title"],
                                "mode_of_issuance": parent["mode_of_issuance"],
                            },
                        }
                    ]
                }
            },
        )

    def _test_delete_mm_document(create_using_pid1=True):
        """Test relation deletion of Multipart Monograph and Document."""

        parent, child = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "DELETE"),
            (
                parent_pid_value,
                parent_pid_type,
                child_pid_value,
                child_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(parent, expected={"relations": {}})
        recrel_assert_record_relations(child, expected={"relations": {}})

    _test_create_mm_document()
    _test_delete_mm_document()
    _test_create_mm_document(create_using_pid1=False)
    _test_delete_mm_document(create_using_pid1=False)
    # recreate for the next one, to have some more valuable test data
    _test_create_mm_document()


def _test_pc_mm_document_with_volume(client, json_headers):
    """Test create and delete relations MM->Document with volume."""

    parent_pid_value = "serid-1"
    parent_pid_type = "serid"
    child_pid_value = "docid-2"
    child_pid_type = "docid"
    relation_type = "multipart_monograph"

    payload = {
        "parent_pid_value": parent_pid_value,
        "parent_pid_type": parent_pid_type,
        "child_pid_value": child_pid_value,
        "child_pid_type": child_pid_type,
        "relation_type": relation_type,
        "volume": "v.3",
    }

    def _test_create_mm_document_with_volume(create_using_pid1=True):
        """Test relation creation with volume of MM and Document."""

        parent, child = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                parent_pid_value,
                parent_pid_type,
                child_pid_value,
                child_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(
            parent,
            expected={
                "relations": {},
            },
        )

        recrel_assert_record_relations(
            child,
            expected={
                "relations": {
                    "multipart_monograph": [
                        {
                            "pid_value": parent_pid_value,
                            "pid_type": parent_pid_type,
                            "volume": "v.3",
                            "relation_type": "multipart_monograph",
                            "record_metadata": {
                                "title": parent["title"],
                                "mode_of_issuance": parent["mode_of_issuance"],
                            },
                        }
                    ]
                },
                "relations_extra_metadata": {
                    "multipart_monograph": [
                        {
                            "pid_value": parent_pid_value,
                            "pid_type": parent_pid_type,
                            "volume": "v.3",
                        }
                    ]
                },
            },
        )

    def _test_delete_mm_document_with_volume(create_using_pid1=True):
        """Test relation deletion with volume of MM and Document."""

        parent, child = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "DELETE"),
            (
                parent_pid_value,
                parent_pid_type,
                child_pid_value,
                child_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(parent, expected={"relations": {}})
        recrel_assert_record_relations(child, expected={"relations": {}})

    _test_create_mm_document_with_volume()
    _test_delete_mm_document_with_volume()
    _test_create_mm_document_with_volume(create_using_pid1=False)
    _test_delete_mm_document_with_volume(create_using_pid1=False)
    # recreate for the next one, to have some more valuable test data
    _test_create_mm_document_with_volume()


def _test_pc_serial_to_mm(client, json_headers):
    """Test relation create/delete of Serial and Multipart Monograph."""
    parent_pid_value = "serid-3"
    parent_pid_type = "serid"
    child_pid_value = "serid-1"
    child_pid_type = "serid"
    relation_type = "serial"

    payload = {
        "parent_pid_value": parent_pid_value,
        "parent_pid_type": parent_pid_type,
        "child_pid_value": child_pid_value,
        "child_pid_type": child_pid_type,
        "relation_type": relation_type,
        "volume": "vol. 1",
    }

    def _test_create_serial_to_mm(create_using_pid1=True):
        """Test relation create of Serial and Multipart Monograph."""

        parent, child = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                parent_pid_value,
                parent_pid_type,
                child_pid_value,
                child_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(
            parent,
            expected={
                "relations": {},
            },
        )

        recrel_assert_record_relations(
            child,
            expected={
                "relations_extra_metadata": {
                    "serial": [
                        {
                            "pid_value": parent_pid_value,
                            "pid_type": parent_pid_type,
                            "volume": "vol. 1",
                        }
                    ]
                },
                "relations": {
                    "serial": [
                        {
                            "pid_value": parent_pid_value,
                            "pid_type": parent_pid_type,
                            "volume": "vol. 1",
                            "relation_type": "serial",
                            "record_metadata": {
                                "title": parent["title"],
                                "mode_of_issuance": parent["mode_of_issuance"],
                            },
                        }
                    ]
                },
            },
        )

    def _test_delete_serial_to_mm(create_using_pid1=True):
        """Test relation delete of Serial and Multipart Monograph."""

        parent, child = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "DELETE"),
            (
                parent_pid_value,
                parent_pid_type,
                child_pid_value,
                child_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(parent, expected={"relations": {}})
        recrel_assert_record_relations(
            child,
            expected={
                "relations": {},
            },
        )

    _test_create_serial_to_mm()
    _test_delete_serial_to_mm()
    _test_create_serial_to_mm(create_using_pid1=False)
    _test_delete_serial_to_mm(create_using_pid1=False)
    # recreate for the next one, to have some more valuable test data
    _test_create_serial_to_mm()


def _test_pc_serial_to_document(client, json_headers):
    """Test relation create/delete of Serial to Document."""
    parent_pid_value = "serid-3"
    parent_pid_type = "serid"
    child_pid_value = "docid-3"
    child_pid_type = "docid"
    relation_type = "serial"

    payload = {
        "parent_pid_value": parent_pid_value,
        "parent_pid_type": parent_pid_type,
        "child_pid_value": child_pid_value,
        "child_pid_type": child_pid_type,
        "relation_type": relation_type,
        "volume": "vol. 1",
    }

    def _test_create_serial_to_document(create_using_pid1=True):
        """Test relation create of Serial to Document."""

        parent, child = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "POST"),
            (
                parent_pid_value,
                parent_pid_type,
                child_pid_value,
                child_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(
            parent,
            expected={
                "relations": {},
            },
        )

        recrel_assert_record_relations(
            child,
            expected={
                "relations": {
                    "serial": [
                        {
                            "pid_value": parent_pid_value,
                            "pid_type": parent_pid_type,
                            "volume": "vol. 1",
                            "relation_type": "serial",
                            "record_metadata": {
                                "title": parent["title"],
                                "mode_of_issuance": parent["mode_of_issuance"],
                            },
                        }
                    ]
                },
                "relations_extra_metadata": {
                    "serial": [
                        {
                            "pid_value": parent_pid_value,
                            "pid_type": parent_pid_type,
                            "volume": "vol. 1",
                        },
                    ]
                },
            },
        )

    def _test_delete_serial_to_document(create_using_pid1=True):
        """Test relation delete of Serial to Document."""

        parent, child = recrel_choose_endpoints_and_do_request(
            (client, json_headers, "DELETE"),
            (
                parent_pid_value,
                parent_pid_type,
                child_pid_value,
                child_pid_type,
            ),
            payload,
            create_using_pid1=create_using_pid1,
        )

        recrel_assert_record_relations(
            parent,
            expected={
                "relations": {},
            },
        )
        recrel_assert_record_relations(child, expected={"relations": {}})

    _test_create_serial_to_document()
    _test_delete_serial_to_document()
    _test_create_serial_to_document(create_using_pid1=False)
    _test_delete_serial_to_document(create_using_pid1=False)
    # recreate for the next one, to have some more valuable test data
    _test_create_serial_to_document()


def _test_pc_invalid_relations_should_fail(
    client, json_headers, invalids, status_code=400
):
    """Test relation creation with invalid parent-child should fail."""
    api_endpoint_documents = "invenio_app_ils_relations.docid_relations"
    api_endpoint_series = "invenio_app_ils_relations.serid_relations"

    for invalid in invalids:
        parent_pid_value = invalid["parent_pid_value"]
        parent_pid_type = invalid["parent_pid_type"]
        child_pid_value = invalid["child_pid_value"]
        child_pid_type = invalid["child_pid_type"]
        relation_type = invalid["relation_type"]

        api_endpoint = (
            api_endpoint_documents
            if parent_pid_type == "docid"
            else api_endpoint_series
        )

        url = url_for(api_endpoint, pid_value=parent_pid_value)
        payload = {
            "parent_pid_value": parent_pid_value,
            "parent_pid_type": parent_pid_type,
            "child_pid_value": child_pid_value,
            "child_pid_type": child_pid_type,
            "relation_type": relation_type,
        }

        res = client.post(url, headers=json_headers, data=json.dumps(payload))
        assert res.status_code == status_code
        if status_code == 400:
            error = json.loads(res.data.decode("utf-8"))
            assert "message" in error
            assert parent_pid_value in error["message"]
            assert child_pid_value in error["message"]


def test_parent_child_relations(client, json_headers, testdata, users):
    """Test parent child relations."""

    _test_pc_invalid_relations_should_fail(
        client,
        json_headers,
        [
            {
                "parent_pid_value": "serid-3",
                "parent_pid_type": "serid",
                "child_pid_value": "docid-1",
                "child_pid_type": "docid",
                "relation_type": "serial",
            }
        ],
        status_code=401,
    )

    user_login(client, "librarian", users)

    # only one test method to speed up tests and avoid testdata recreation at
    # each test. As drawback, testdata is not cleaned between each test, so
    # do not change the order of execution of the following tests :)

    # serid-1 [MM] -> docid-1
    _test_pc_mm_document(client, json_headers)

    #      serid-1 [MM]
    #          |
    #        docid-1

    # serid-1 [MM] --- v.3 ---> docid-2
    _test_pc_mm_document_with_volume(client, json_headers)

    #         serid-1 [MM]
    #             |
    #            / \ v.3
    #     docid-1   docid-2

    # serid-3 [SERIAL] -> serid-1 [MM]
    _test_pc_serial_to_mm(client, json_headers)

    #        serid-3 [SERIAL]
    #             |
    #         serid-1 [MM]
    #              |
    #             / \ v.3
    #      docid-1   docid-2

    # serid-3 [SERIAL] -> docid-3
    _test_pc_serial_to_document(client, json_headers)

    #           serid-3 [SERIAL]
    #                  |
    #                 / \
    #          docid-3   serid-1 [MM]
    #                         |
    #                        / \ v.3
    #                 docid-1   docid-2

    # test wrong relations
    invalids = [
        {
            "parent_pid_value": "docid-1",
            "parent_pid_type": "docid",
            "child_pid_value": "serid-1",  # Multipart Monograph
            "child_pid_type": "serid",
            "relation_type": "multipart_monograph",
        },
        {
            "parent_pid_value": "docid-1",
            "parent_pid_type": "docid",
            "child_pid_value": "serid-3",  # Serial
            "child_pid_type": "serid",
            "relation_type": "serial",
        },
        {
            "parent_pid_value": "serid-1",  # Multipart Monograph
            "parent_pid_type": "serid",
            "child_pid_value": "serid-3",  # Serial
            "child_pid_type": "serid",
            "relation_type": "multipart_monograph",
        },
        {
            "parent_pid_value": "docid-1",
            "parent_pid_type": "docid",
            "child_pid_value": "docid-2",
            "child_pid_type": "docid",
            "relation_type": "serial",
        },
        # already existing
        {
            "parent_pid_value": "serid-3",  # Serial
            "parent_pid_type": "serid",
            "child_pid_value": "docid-3",
            "child_pid_type": "docid",
            "relation_type": "serial",
        },
    ]
    _test_pc_invalid_relations_should_fail(client, json_headers, invalids)


def test_max_one_multipart(testdata):
    doc = testdata["documents"][0]
    series1 = testdata["series"][0]
    rr = RecordRelationsParentChild()
    res = rr.add(
        series1,
        doc,
        Relation.get_relation_by_name("multipart_monograph"),
    )

    assert res

    series2 = testdata["series"][1]

    with pytest.raises(RecordRelationsError) as error:
        rr = RecordRelationsParentChild()
        res = rr.add(
            series2,
            doc,
            "multipart_monograph",
        )

    assert isinstance(error.value, RecordRelationsError)
